import random
from collections.abc import Mapping, Sequence
from typing import Any
from uuid import UUID, uuid4

from haiway import Meta, State, as_dict, ctx, execute_concurrently

from draive.evaluation import (
    EvaluatorSuiteCase,
    EvaluatorSuiteResult,
    PreparedEvaluatorSuite,
)
from draive.models import Instructions, InstructionsRepository, ModelInstructions
from draive.multimodal import MultimodalContent, MultimodalTag, TextContent
from draive.parameters import DataModel
from draive.stages import Stage, StageState, stage

__all__ = ("refine_instructions",)


async def refine_instructions[Parameters: DataModel](
    instructions: Instructions,
    /,
    *,
    instructions_content: ModelInstructions | None = None,
    guidelines: str | None = None,
    evaluator_suite: PreparedEvaluatorSuite[Parameters],
    evaluator_cases: Sequence[EvaluatorSuiteCase[Parameters]],
    rounds_limit: int,
    sample_ratio: float = 0.1,
    candidates_limit: int = 3,
    performance_drop_threshold: float = 0.5,
    quality_threshold: float = 0.99,
    concurrent_nodes: int = 1,
) -> ModelInstructions:
    """
    Refine instructions using binary tree exploration with performance pruning.

    Each node generates exactly 2 complementary strategies.
    Branches are pruned if performance drops below threshold.

    Args:
        instructions: The instructions to refine
        guidelines: Optional guidelines for refinement
        evaluator_suite: Suite to evaluate instructions performance
        rounds_limit: Maximum depth of refinement tree
        sample_ratio: Fraction of passing cases to include in focused evaluation
        candidates_limit: Number of top candidates to fully evaluate
        performance_drop_threshold: Prune branches with score drop > this threshold
        quality_threshold: Stop if score reaches this threshold
        concurrent_nodes: How many nodes explored concurrently
    """

    assert rounds_limit > 0  # nosec: B101
    assert 1 >= sample_ratio > 0  # nosec: B101
    assert candidates_limit > 0  # nosec: B101
    assert 1 >= performance_drop_threshold > 0  # nosec: B101
    assert 1 >= quality_threshold >= 0  # nosec: B101
    assert concurrent_nodes > 0  # nosec: B101
    assert len(evaluator_cases) > 0  # nosec: B101

    ctx.log_info(
        f"Starting tree-based refinement: "
        f"max_depth={rounds_limit}, "
        f"sample_ratio={sample_ratio}, "
        f"performance_drop_threshold={performance_drop_threshold}"
    )

    result: MultimodalContent = await Stage.sequence(
        _tree_initialization_stage(
            instructions=instructions,
            instructions_content=instructions_content,
            evaluator_suite=evaluator_suite,
            sample_ratio=sample_ratio,
            performance_drop_threshold=performance_drop_threshold,
            guidelines=guidelines,
            rounds_limit=rounds_limit,
        ),
        _tree_exploration_stage(
            evaluator_suite=evaluator_suite,
            evaluator_cases=evaluator_cases,
            rounds_limit=rounds_limit,
            quality_threshold=quality_threshold,
            concurrent_nodes=concurrent_nodes,
        ),
        _tree_finalization_stage(
            evaluator_suite=evaluator_suite,
            quality_threshold=quality_threshold,
            candidates_limit=candidates_limit,
        ),
    ).execute()

    ctx.log_info("...instructions refinement finished!")
    return result.to_str()


class _RefinementTreeNode(State):
    identifier: UUID
    instructions: Instructions
    instructions_content: ModelInstructions
    strategy: str
    parent_id: UUID | None
    depth: int
    focused_evaluation: EvaluatorSuiteResult
    complete_evaluation: EvaluatorSuiteResult | None
    children: Sequence[UUID]
    pruned: bool

    @property
    def is_root(self) -> bool:
        return self.parent_id is None

    @property
    def is_leaf(self) -> bool:
        return len(self.children) == 0

    @property
    def focused_evaluation_performance(self) -> float:
        return self.focused_evaluation.performance

    @property
    def complete_evaluation_performance(self) -> float | None:
        if self.complete_evaluation is None:
            return None

        return self.complete_evaluation.performance

    @property
    def patched_instructions_repository(self) -> InstructionsRepository:
        repository: InstructionsRepository = ctx.state(InstructionsRepository)

        async def instructions_loading(
            name: str,
            meta: Meta,
            **extra: Any,
        ) -> str | None:
            # use updated instructions if able
            if name == self.instructions.name:
                return self.instructions_content

            else:  # otherwise preserve unchanged instructions
                return await repository.loading(
                    name=name,
                    meta=meta,
                    **extra,
                )

        return InstructionsRepository(
            listing=repository.listing,  # keep current listing
            loading=instructions_loading,  # replace loading
            # do not allow modifications - use default noop implementation
        )


class _RefinementState(State):
    root: _RefinementTreeNode
    nodes: Mapping[UUID, _RefinementTreeNode]
    sample_ratio: float
    performance_drop_threshold: float
    guidelines: str | None = None
    rounds_remaining: int

    def leaves(self) -> Sequence[_RefinementTreeNode]:
        return [node for node in self.nodes.values() if node.is_leaf and not node.pruned]


def _select_focused_cases[Parameters: DataModel](
    *,
    evaluation_result: EvaluatorSuiteResult,
    evaluator_cases: Sequence[EvaluatorSuiteCase[Parameters]],
    sample_ratio: float,
) -> Sequence[str]:
    # Get all failing cases
    failing_cases: Sequence[str] = [
        case_result.case_identifier
        for case_result in evaluation_result.results
        if not case_result.passed
    ]

    # Get passing cases
    passing_cases: Sequence[str] = [
        case_result.case_identifier
        for case_result in evaluation_result.results
        if case_result.passed
    ]

    # Get other, previously excluded cases
    additional_cases: Sequence[str] = [
        case.identifier
        for case in evaluator_cases
        if case.identifier not in failing_cases and case.identifier not in passing_cases
    ]

    sampling_cases_pool: Sequence[str] = passing_cases + additional_cases
    sample_size: int = (
        max(1, int(len(sampling_cases_pool) * sample_ratio)) if sampling_cases_pool else 0
    )
    sampling_cases: Sequence[str] = (
        random.sample(  # nosec: B311
            sampling_cases_pool,
            min(len(sampling_cases_pool), sample_size),
        )
        if sample_size > 0
        else []
    )

    ctx.log_info(
        f"Created focused evaluation suite: "
        f"{len(failing_cases)} failing + {len(sampling_cases)} sampled other "
        f"(out of {len(passing_cases)} total passing)"
    )

    # Combine for focused evaluation
    return failing_cases + sampling_cases


def _tree_initialization_stage[Parameters: DataModel](
    *,
    instructions: Instructions,
    instructions_content: ModelInstructions | None,
    evaluator_suite: PreparedEvaluatorSuite[Parameters],
    sample_ratio: float,
    performance_drop_threshold: float,
    guidelines: str | None,
    rounds_limit: int,
) -> Stage:
    @stage
    async def tree_initialization(
        *,
        state: StageState,
    ) -> StageState:
        ctx.log_info("...evaluating initial instructions with full suite...")
        # Make initial evaluation using current instructions
        evaluation: EvaluatorSuiteResult = await evaluator_suite()

        ctx.log_info(f"...initial score: {evaluation.performance:.2f}...")

        content: ModelInstructions
        if instructions_content is None:
            content = await InstructionsRepository.load(instructions)

        else:
            content = instructions_content

        # Create root node
        root_node = _RefinementTreeNode(
            identifier=uuid4(),
            instructions=instructions,
            instructions_content=content,
            strategy="initial",
            parent_id=None,
            depth=0,
            children=(),
            # for root complete evaluation is the same as focused
            focused_evaluation=evaluation,
            complete_evaluation=evaluation,
            pruned=False,
        )

        # Setup initial state
        return state.updated(
            _RefinementState(
                root=root_node,
                nodes={root_node.identifier: root_node},
                sample_ratio=sample_ratio,
                performance_drop_threshold=performance_drop_threshold,
                guidelines=guidelines,
                rounds_remaining=rounds_limit,
            ),
            # set the result to be initial instruction content
            result=MultimodalContent.of(content),
        )

    return tree_initialization


def _tree_exploration_stage[Parameters: DataModel](
    *,
    evaluator_suite: PreparedEvaluatorSuite[Parameters],
    evaluator_cases: Sequence[EvaluatorSuiteCase[Parameters]],
    rounds_limit: int,
    quality_threshold: float,
    concurrent_nodes: int,
) -> Stage:
    @stage
    async def tree_exploration(
        *,
        state: StageState,
    ) -> StageState:
        refinement_state: _RefinementState = state.get(
            _RefinementState,
            required=True,
        )

        async def explore(
            node: _RefinementTreeNode,
        ) -> Mapping[UUID, _RefinementTreeNode]:
            return await _explore_node(
                node=node,
                evaluator_suite=evaluator_suite,
                evaluator_cases=evaluator_cases,
                sample_ratio=refinement_state.sample_ratio,
                performance_drop_threshold=refinement_state.performance_drop_threshold,
                guidelines=refinement_state.guidelines,
            )

        node_updates: Sequence[Mapping[UUID, _RefinementTreeNode]] = await execute_concurrently(
            explore,
            refinement_state.leaves(),
            concurrent_tasks=concurrent_nodes,
        )

        updated_nodes: dict[UUID, _RefinementTreeNode] = as_dict(refinement_state.nodes)
        for nodes in node_updates:
            updated_nodes = {
                **updated_nodes,
                **nodes,
            }

        # Update the refinement state with the updated nodes
        refinement_state = refinement_state.updated(nodes=updated_nodes)

        # Log tree statistics
        total_nodes: int = len(refinement_state.nodes)
        pruned_count: int = len([node for node in refinement_state.nodes.values() if node.pruned])
        active_nodes: int = total_nodes - pruned_count

        ctx.log_info(
            f"Tree exploration round complete. Nodes: {total_nodes} total, "
            f"{active_nodes} active, {pruned_count} pruned"
        )

        return state.updated(refinement_state)

    async def condition(
        *,
        state: StageState,
        iteration: int,
    ) -> bool:
        refinement_state: _RefinementState = state.get(
            _RefinementState,
            required=True,
        )
        if any(
            (
                node.complete_evaluation_performance
                if node.complete_evaluation_performance is not None
                else node.focused_evaluation_performance
            )
            > quality_threshold
            for node in refinement_state.nodes.values()
        ):
            # we can complete early with exceptional score
            return False

        return iteration < rounds_limit

    return Stage.loop(
        tree_exploration,
        condition=condition,
    )


async def _explore_node[Parameters: DataModel](
    node: _RefinementTreeNode,
    evaluator_suite: PreparedEvaluatorSuite[Parameters],
    evaluator_cases: Sequence[EvaluatorSuiteCase[Parameters]],
    sample_ratio: float,
    performance_drop_threshold: float,
    guidelines: str | None,
) -> Mapping[UUID, _RefinementTreeNode]:
    assert not node.pruned  # nosec: B101 # skip pruned branches
    ctx.log_info(
        f"Exploring node {node.identifier} at"
        f" depth {node.depth}, score: {node.focused_evaluation_performance:.2f}"
    )

    # Generate exactly 2 complementary strategies
    strategies: Sequence[tuple[str, ModelInstructions]] = await _generate_refined_instructions(
        current_content=node.instructions_content,
        evaluation_result=node.focused_evaluation,
        parent_strategy=node.strategy,
        guidelines=guidelines,
    )

    # Create and evaluate child nodes
    focused_suite_cases: Sequence[str] = _select_focused_cases(
        evaluation_result=node.focused_evaluation,
        evaluator_cases=evaluator_cases,
        sample_ratio=sample_ratio,
    )

    children: dict[UUID, _RefinementTreeNode] = {}
    for strategy_name, refined_instructions in strategies:
        # Create child node
        child_node = _RefinementTreeNode(
            identifier=uuid4(),
            instructions=node.instructions,
            instructions_content=refined_instructions,
            strategy=strategy_name,
            parent_id=node.identifier,
            depth=node.depth + 1,
            # initialize with placeholder
            focused_evaluation=EvaluatorSuiteResult(
                suite="placeholder",
                results=(),
            ),
            complete_evaluation=None,
            children=(),
            pruned=False,
        )

        # Evaluate with focused suite
        ctx.log_info(f"Evaluating strategy '{strategy_name}'...")
        focused_evaluation: EvaluatorSuiteResult
        with ctx.updated(child_node.patched_instructions_repository):
            focused_evaluation = await evaluator_suite(focused_suite_cases)

        # Check for performance drop
        performance_ratio = (
            focused_evaluation.performance / node.focused_evaluation_performance
            if node.focused_evaluation_performance > 0
            else 0
        )
        pruned: bool = performance_ratio < performance_drop_threshold

        # update node with evaluation data
        children[child_node.identifier] = child_node.updated(
            focused_evaluation=focused_evaluation,
            pruned=pruned,
        )

        if pruned:
            ctx.log_warning(
                f"Strategy '{strategy_name}' caused {(1 - performance_ratio) * 100:.1f}% "
                f"performance drop. Pruning this branch."
            )

        else:
            ctx.log_info(
                f"Created child node {child_node.identifier}: strategy={strategy_name}, "
                f"score={focused_evaluation.performance:.2f}"
                f" ({performance_ratio:.1%} of parent)"
            )

    return children


async def _generate_strategy_metadata(
    *,
    evaluation_result: EvaluatorSuiteResult,
    parent_strategy: str | None,
    guidelines: str | None,
) -> Sequence[tuple[str, str]]:
    """
    Generate strategy metadata (name and approach) without instructions content.

    Returns:
        Sequence of tuples containing (name, approach) for each strategy.
    """
    # Analyze failures
    failure_report = evaluation_result.report(
        detailed=True,
        include_passed=False,
    )

    strategy_prompt: str = f"""
You are an expert prompt engineer providing feedback and recommendations for instructions improvement.
Based on the evaluation failures analysis, prepare EXACTLY 2 DIFFERENT refinement strategies.

<previous_strategy>{parent_strategy or "Initial"}</previous_strategy>
<task>
Generate 2 instructions refinement strategies and recommendations that:
1. Take CONTRASTING approaches to fixing the issues
2. Address different aspects of the failures
3. Are mutually exclusive in their approach
</task>
{f"\n<guidelines>\n{guidelines}</guidelines>\n" if guidelines else ""}
<examples>
<strategy>
<name>Exemplification</name>
<approach>Add more specific details and examples</approach>
</strategy>
<strategy>
<name>Constraining</name>
<approach>Add explicit constraints and rules</approach>
</strategy>
</examples>
<format>
For each strategy, provide the name and approach in the following tags format:
<strategy>
<name>[Brief descriptive name]</name>
<approach>[Explanation of the approach with description of recommendations]</approach>
</strategy>
</format>
"""  # noqa: E501

    response: MultimodalContent = await Stage.completion(
        f"<failure_analysis>\n{failure_report}\n</failure_analysis>",
        instructions=strategy_prompt,
    ).execute()

    # Parse strategies
    strategies: list[tuple[str, str]] = []
    for strategy_element in response.tags("strategy"):
        name_element: MultimodalTag | None = strategy_element.content.tag("name")
        approach_element: MultimodalTag | None = strategy_element.content.tag("approach")

        if name_element and approach_element:
            strategy_name: str = name_element.content.to_str().strip()
            strategy_approach: str = approach_element.content.to_str().strip()
            strategies.append((strategy_name, strategy_approach))

    return strategies


async def _generate_instructions(
    *,
    current_content: ModelInstructions,
    strategy_name: str,
    strategy_approach: str,
    evaluation_result: EvaluatorSuiteResult,
    guidelines: str | None,
) -> ModelInstructions:
    # Analyze failures
    failure_report: str = evaluation_result.report(
        detailed=True,
        include_passed=False,
    )

    refinement_prompt: str = f"""\
You are an expert prompt engineer refining provided instructions referred as a subject.
Provide an updated version of the subject that implements the described strategy to address the evaluation failures.
Strictly focus on delivering the best possible instructions content with excellent clarity and performance.
Ensure to apply modifications by following requested strategy while keeping original instructions intent, goal and formatting style.

<strategy>
<name>
{strategy_name}
</name>
<approach>
{strategy_approach}
</approach>
</strategy>

<subject>
{current_content}
</subject>
{f"\n<guidelines>\n{guidelines}</guidelines>\n" if guidelines else ""}
<format>
Provide ONLY the refined instructions content, without any explanation or metadata.
</format>
"""  # noqa: E501

    ctx.log_info(f"Generating updated instructions using {strategy_name}:\n\n{strategy_approach}")
    response: MultimodalContent = await Stage.completion(
        f"<failure_analysis>\n{failure_report}\n</failure_analysis>",
        instructions=refinement_prompt,
    ).execute()

    updated_instruction: str = response.to_str().strip()

    ctx.log_info(f"Prepared updated instructions using {strategy_name}")
    ctx.log_debug(f"Updated instruction for {strategy_name}: {updated_instruction}")

    return updated_instruction


async def _generate_refined_instructions(
    *,
    current_content: ModelInstructions,
    evaluation_result: EvaluatorSuiteResult,
    parent_strategy: str | None,
    guidelines: str | None,
) -> Sequence[tuple[str, ModelInstructions]]:
    # Step 1: Generate strategy metadata
    strategy_metadata: Sequence[tuple[str, str]] = await _generate_strategy_metadata(
        evaluation_result=evaluation_result,
        parent_strategy=parent_strategy,
        guidelines=guidelines,
    )

    # Step 2: Generate instructions content for each strategy
    refined_strategies: list[tuple[str, ModelInstructions]] = []
    for strategy_name, strategy_approach in strategy_metadata:
        refined_instructions = await _generate_instructions(
            current_content=current_content,
            strategy_name=strategy_name,
            strategy_approach=strategy_approach,
            evaluation_result=evaluation_result,
            guidelines=guidelines,
        )
        refined_strategies.append((strategy_name, refined_instructions))

    # Validate and return results
    match len(refined_strategies):
        case 2:
            return refined_strategies

        case 1:
            ctx.log_warning(f"Expected 2 strategies, got {len(refined_strategies)}")
            return refined_strategies

        case 0:
            ctx.log_warning("Failed to generate refinement strategies")
            return refined_strategies  # this branch will be skipped

        case _:  # more than 2
            ctx.log_warning(f"Expected 2 strategies, got {len(refined_strategies)}")
            return refined_strategies[:2]


def _tree_finalization_stage[Parameters: DataModel](
    *,
    evaluator_suite: PreparedEvaluatorSuite[Parameters],
    quality_threshold: float,
    candidates_limit: int,
) -> Stage:
    @stage
    async def tree_finalization(
        *,
        state: StageState,
    ) -> StageState:
        refinement_state: _RefinementState = state.get(
            _RefinementState,
            required=True,
        )
        # Find best candidates across entire tree (except root)
        candidates: Sequence[_RefinementTreeNode] = _find_best_candidates(
            refinement_state=refinement_state,
            quality_threshold=quality_threshold,
            limit=candidates_limit,
        )

        if not candidates:
            ctx.log_warning("No candidates found, keeping original instruction")
            return state

        ctx.log_info(f"Running full evaluation on top {len(candidates)} candidates")

        # Full evaluation of top candidates
        best_instructions: ModelInstructions = refinement_state.root.instructions_content
        best_score: float = refinement_state.root.complete_evaluation_performance or 0
        best_node: _RefinementTreeNode = refinement_state.root

        for candidate_node in candidates:
            ctx.log_info(
                f"Full evaluation of node {candidate_node.identifier} "
                f"(strategy: {candidate_node.strategy}, depth: {candidate_node.depth})"
            )

            complete_evaluation: EvaluatorSuiteResult
            with ctx.updated(candidate_node.patched_instructions_repository):
                complete_evaluation = await evaluator_suite()

            # Update node with full eval score
            updated_node: _RefinementTreeNode = candidate_node.updated(
                complete_evaluation=complete_evaluation
            )
            refinement_state = refinement_state.updated(
                nodes={
                    **refinement_state.nodes,
                    # update node in tree
                    updated_node.identifier: updated_node,
                },
            )

            ctx.log_info(
                f"Full evaluation score: {complete_evaluation.performance:.4f} "
                f"(focused was: {updated_node.focused_evaluation_performance:.4f})"
            )

            if complete_evaluation.performance > best_score:
                best_instructions = updated_node.instructions_content
                best_score = complete_evaluation.performance
                best_node = updated_node

        # Log final statistics
        _log_tree_statistics(
            refinement_state=refinement_state,
            best_node=best_node,
        )

        # Update result with best instructions and updated state
        return state.updated(
            refinement_state,
            result=MultimodalContent.of(
                TextContent.of(
                    best_instructions,
                    meta={"score": best_score},
                ),
            ),
        )

    return tree_finalization


def _find_best_candidates(
    *,
    refinement_state: _RefinementState,
    quality_threshold: float,
    limit: int,
) -> Sequence[_RefinementTreeNode]:
    # Get all nodes that are not pruned
    candidates: list[_RefinementTreeNode] = []
    for node in refinement_state.nodes.values():
        # skip excluded and root from candidates
        if node.pruned or node.is_root:
            continue

        # Include if it's a leaf or has exceptional performance
        if node.is_leaf or node.focused_evaluation_performance > quality_threshold:
            candidates.append(node)

    # Sort by focused score and depth (prefer deeper nodes with same score)
    candidates.sort(
        key=lambda x: (x.focused_evaluation_performance, x.depth),
        reverse=True,
    )

    return candidates[:limit]


def _log_tree_statistics(
    *,
    refinement_state: _RefinementState,
    best_node: _RefinementTreeNode,
) -> None:
    total_nodes: int = len(refinement_state.nodes)
    pruned_count: int = len([node for node in refinement_state.nodes.values() if node.pruned])
    active_nodes: int = total_nodes - pruned_count

    # Depth distribution
    depth_distribution: dict[int, int] = {}
    for node in refinement_state.nodes.values():
        depth_distribution[node.depth] = depth_distribution.get(node.depth, 0) + 1

    max_depth: int = max(depth_distribution.keys()) if depth_distribution else 0

    # Build path to best node
    best_path: list[str] = []
    current_id: UUID | None = best_node.identifier
    while current_id:
        node = refinement_state.nodes[current_id]
        best_path.append(f"{node.strategy} (score: {node.focused_evaluation_performance:.3f})")
        current_id = node.parent_id

    best_path.reverse()

    ctx.log_info(
        f"Tree statistics:\n"
        f"- Total nodes explored: {total_nodes}\n"
        f"- Active nodes: {active_nodes}\n"
        f"- Pruned nodes: {pruned_count} ({pruned_count / total_nodes * 100:.1f}%)\n"
        f"- Max depth reached: {max_depth}\n"
        f"- Best path: {' -> '.join(best_path) if best_path else 'None'}"
    )



import re
from base64 import b64decode
from collections.abc import Callable, Iterable, MutableMapping, Sequence
from datetime import UTC, datetime, timedelta
from typing import Any, cast

from haiway import AttributePath, AttributeRequirement, as_tuple, ctx
from haiway.postgres import Postgres, PostgresRow, PostgresValue

from draive.embedding import (
    Embedded,
    ImageEmbedding,
    TextEmbedding,
    mmr_vector_similarity_search,
)
from draive.multimodal import TextContent
from draive.parameters import DataModel
from draive.resources import ResourceContent
from draive.utils import VectorIndex

__all__ = ("PostgresVectorIndex",)


# POSTGRES SCHEMA TEMPLATE
#
# CREATE TABLE IF NOT EXISTS your_table_name (
#     id UUID NOT NULL DEFAULT gen_random_uuid(),
#     embedding VECTOR(<dimension>) NOT NULL,
#     payload JSONB NOT NULL,
#     meta JSONB NOT NULL,
#     created TIMESTAMPTZ NOT NULL DEFAULT CURRENT_TIMESTAMP
# );
#
# -- Recommended for cosine similarity
# CREATE INDEX IF NOT EXISTS your_table_name_embedding_idx
#     ON your_table_name
#     USING ivfflat (embedding vector_cosine_ops)
#     WITH (lists = 100);
#


def PostgresVectorIndex(  # noqa: C901, PLR0915
    *,
    mmr_multiplier: int = 8,
) -> VectorIndex:
    """Create a Postgres-backed implementation of :class:`VectorIndex`.

    Parameters
    ----------
    mmr_multiplier
        Multiplier applied to ``limit`` to determine how many database rows are
        fetched before applying Maximal Marginal Relevance (MMR) re-ranking.

    Returns
    -------
    VectorIndex
        A VectorIndex implementation persisting entries in Postgres.
    """
    assert mmr_multiplier > 0  # nosec: B101

    table_names: MutableMapping[type[DataModel], str] = {}

    def resolve_table_name(model: type[DataModel]) -> str:
        nonlocal table_names
        if name := table_names.get(model):
            return name

        resolved: str = re.sub(r"(?<!^)(?=[A-Z])", "_", model.__name__).lower()
        table_names[model] = resolved
        return resolved

    async def index[Model: DataModel, Value: ResourceContent | TextContent | str](
        model: type[Model],
        /,
        *,
        attribute: Callable[[Model], Value] | AttributePath[Model, Value] | Value,
        values: Iterable[Model],
        **extra: Any,
    ) -> None:
        value_selector: Callable[[Model], Value]
        match attribute:
            case Callable() as selector:
                value_selector = selector

            case path:
                assert isinstance(  # nosec: B101
                    path, AttributePath
                ), "Prepare parameter path by using Self._.path.to.property"
                value_selector = cast(AttributePath[Model, Value], path).__call__

        value_models: Sequence[Model] = as_tuple(values)
        selected_values: list[str | bytes] = []
        for value in value_models:
            match value_selector(value):
                case str() as text:
                    selected_values.append(text)

                case TextContent() as text_content:
                    selected_values.append(text_content.text)

                case ResourceContent() as resource_content:
                    if not resource_content.mime_type.startswith("image"):
                        raise ValueError(f"{resource_content.mime_type} embedding is not supported")

                    selected_values.append(b64decode(resource_content.data))

                case other:
                    # we could download items from resource references for embedding if needed
                    raise ValueError(f"{other} embedding is not supported")

        embedded_values: Sequence[Embedded[Model]]
        if all(isinstance(value, str) for value in selected_values):
            embedded_values = [
                Embedded(
                    value=value,
                    vector=embedded.vector,
                    meta=embedded.meta,
                )
                for embedded, value in zip(
                    await TextEmbedding.embed(
                        cast(list[str], selected_values),
                        **extra,
                    ),
                    value_models,
                    strict=True,
                )
            ]

        elif all(isinstance(value, bytes) for value in selected_values):
            embedded_values = [
                Embedded(
                    value=value,
                    vector=embedded.vector,
                    meta=embedded.meta,
                )
                for embedded, value in zip(
                    await ImageEmbedding.embed(
                        cast(list[bytes], selected_values),
                        **extra,
                    ),
                    value_models,
                    strict=True,
                )
            ]

        else:
            raise ValueError("Selected attribute values have to be the same type")

        created_timestamp: datetime = datetime.now(UTC)
        async with Postgres.acquire_connection() as connection:
            async with connection.transaction():
                for idx, embedded in enumerate(embedded_values):
                    await connection.execute(
                        f"""
                        INSERT INTO {resolve_table_name(model)} (
                            embedding,
                            payload,
                            meta,
                            created
                        )

                        VALUES (
                            $1::VECTOR,
                            $2::JSONB,
                            $3::JSONB,
                            $4::TIMESTAMPTZ
                        );
                        """,  # nosec: B608
                        embedded.vector,
                        embedded.value.to_json(),
                        embedded.meta.to_json(),
                        created_timestamp + timedelta(microseconds=idx),
                    )

        ctx.log_info("Vector index update completed.")

    async def search[Model: DataModel](  # noqa: C901, PLR0912
        model: type[Model],
        /,
        *,
        query: Sequence[float] | ResourceContent | TextContent | str | None = None,
        score_threshold: float | None = None,
        requirements: AttributeRequirement[Model] | None = None,
        limit: int | None = None,
        rerank: bool = False,
        **extra: Any,
    ) -> Sequence[Model]:
        assert query is not None or (query is None and score_threshold is None)  # nosec: B101
        where_clause: str
        arguments: Sequence[Sequence[PostgresValue] | PostgresValue]
        if query is None:
            where_clause, arguments = resolve_requirements(requirements)
            if where_clause:
                where_clause = f"WHERE {where_clause}"

            parameters: Sequence[Sequence[PostgresValue] | PostgresValue] = [
                *arguments,
                limit or 8,
            ]
            results: Sequence[PostgresRow] = await Postgres.fetch(
                f"""
                SELECT
                    payload

                FROM {resolve_table_name(model)}

                {where_clause}
                ORDER BY created DESC
                LIMIT ${len(parameters)};
                """,  # nosec: B608
                *parameters,
            )

            return tuple(model.from_json(cast(str, result["payload"])) for result in results)

        query_vector: Sequence[float]
        match query:
            case str() as text:
                embedded_query: Embedded[str] = await TextEmbedding.embed(text)
                query_vector = embedded_query.vector

            case TextContent() as text_content:
                embedded_query: Embedded[str] = await TextEmbedding.embed(text_content.text)
                query_vector = embedded_query.vector

            case ResourceContent() as resource_content:
                if resource_content.mime_type.startswith("image"):
                    embedded_image: Embedded[bytes] = await ImageEmbedding.embed(
                        b64decode(resource_content.data)
                    )
                    query_vector = embedded_image.vector

                elif resource_content.mime_type.startswith("text"):
                    embedded_query: Embedded[str] = await TextEmbedding.embed(
                        b64decode(resource_content.data).decode()
                    )
                    query_vector = embedded_query.vector

                else:
                    raise ValueError(f"{resource_content.mime_type} embedding is not supported")

            case vector:
                query_vector = vector

        arguments: Sequence[Sequence[PostgresValue] | PostgresValue] = (query_vector,)
        similarity_expression: str = f"embedding <#> ${len(arguments)}"

        where_clause, arguments = resolve_requirements(requirements, arguments=arguments)

        if score_threshold is not None:
            arguments = (*arguments, 1.0 - float(score_threshold))
            threshold_clause: str = f"{similarity_expression} <= ${len(arguments)}"
            if where_clause:
                where_clause = f"WHERE {threshold_clause} AND ({where_clause})"

            else:
                where_clause = f"WHERE {threshold_clause}"

        elif where_clause:
            where_clause = f"WHERE {where_clause}"

        arguments = (*arguments, (limit or 8) * mmr_multiplier if rerank else (limit or 8))
        results: Sequence[PostgresRow] = await Postgres.fetch(
            f"""
            SELECT
                embedding,
                payload

            FROM {resolve_table_name(model)}

            {where_clause}
            ORDER BY {similarity_expression}
            LIMIT ${len(arguments)};
            """,  # nosec: B608
            *arguments,
        )

        if not rerank:
            return tuple(model.from_json(cast(str, result["payload"])) for result in results)

        matching: list[Embedded[Model]] = [
            Embedded[Model](
                vector=cast(Sequence[float], result["embedding"]),
                value=model.from_json(cast(str, result["payload"])),
            )
            for result in results
        ]
        return tuple(
            matching[index].value
            for index in mmr_vector_similarity_search(
                query_vector=query_vector,
                values_vectors=[element.vector for element in matching],
                limit=limit,
            )
        )

    async def delete[Model: DataModel](
        model: type[Model],
        /,
        *,
        requirements: AttributeRequirement[Model] | None = None,
        **extra: Any,
    ) -> None:
        if requirements is None:
            await Postgres.execute(
                f"""
                DELETE FROM {resolve_table_name(model)};
                """,  # nosec: B608
            )
            ctx.log_info(f"Removed all entries for {model.__name__}.")

            return

        where_clause: str
        arguments: Sequence[Sequence[PostgresValue] | PostgresValue]
        where_clause, arguments = resolve_requirements(requirements)
        if where_clause:
            where_clause = f"WHERE {where_clause}"

        await Postgres.execute(
            f"""
            DELETE FROM {resolve_table_name(model)}
            {where_clause};
            """,  # nosec: B608
            *arguments,
        )
        ctx.log_info(f"Removed filtered entries for {model.__name__}.")

    return VectorIndex(
        indexing=index,
        searching=search,
        deleting=delete,
    )


def _resolve_requirement(  # noqa: PLR0911
    requirement: AttributeRequirement[Any],
    /,
    arguments: Sequence[Sequence[PostgresValue] | PostgresValue],
) -> tuple[str, Sequence[Sequence[PostgresValue] | PostgresValue]]:
    resolved_arguments: Sequence[Sequence[PostgresValue] | PostgresValue]
    match requirement.operator:
        case "and":
            left_sql, partial_arguments = _resolve_requirement(
                requirement.lhs,
                arguments=arguments,
            )
            right_sql, resolved_arguments = _resolve_requirement(
                requirement.rhs,
                arguments=partial_arguments,
            )
            return f"({left_sql} AND {right_sql})", resolved_arguments

        case "or":
            left_sql, partial_arguments = _resolve_requirement(
                requirement.lhs,
                arguments=arguments,
            )
            right_sql, resolved_arguments = _resolve_requirement(
                requirement.rhs,
                arguments=partial_arguments,
            )
            return f"({left_sql} OR {right_sql})", resolved_arguments

        case "equal":
            resolved_arguments = [*arguments, requirement.rhs]
            return (
                f"{_scalar_accessor(str(requirement.lhs))} = ${len(resolved_arguments)}",
                resolved_arguments,
            )

        case "not_equal":
            resolved_arguments = [*arguments, requirement.rhs]
            return (
                f"({_scalar_accessor(str(requirement.lhs))}"
                f" IS DISTINCT FROM ${len(resolved_arguments)})",
                resolved_arguments,
            )

        case "contained_in":
            resolved_arguments = [*arguments, requirement.rhs]
            return (
                f"{_scalar_accessor(str(requirement.lhs))} = ANY(${len(resolved_arguments)})",
                resolved_arguments,
            )

        case "contains_any":
            resolved_arguments = [*arguments, requirement.rhs]
            return (
                "EXISTS (SELECT 1 FROM jsonb_array_elements_text("  # nosec: B608
                f"{_scalar_accessor(str(requirement.lhs))}) AS element"
                f" WHERE element = ANY(${len(resolved_arguments)}))",
                resolved_arguments,
            )

        case "contains":
            resolved_arguments = [*arguments, requirement.rhs]
            return (
                "EXISTS (SELECT 1 FROM jsonb_array_elements_text("  # nosec: B608
                f"{_scalar_accessor(str(requirement.lhs))} AS element"
                f" WHERE element = ${len(resolved_arguments)})",
                resolved_arguments,
            )

        case "text_match":
            # TODO: text match LIKE
            raise NotImplementedError("Not implemented yet")


def resolve_requirements(
    requirement: AttributeRequirement[Any] | None,
    /,
    arguments: Sequence[Sequence[PostgresValue] | PostgresValue] = (),
) -> tuple[str, Sequence[Sequence[PostgresValue] | PostgresValue]]:
    if requirement is None:
        return ("", arguments)

    where_clause: str
    where_clause, arguments = _resolve_requirement(
        requirement,
        arguments=arguments,
    )

    return (where_clause, arguments)


def _path_literal(path: str) -> str:
    return f"'{{{','.join(path.lstrip('.').split('.'))}}}'"


def _scalar_accessor(path: str) -> str:
    return f"payload #>> {_path_literal(path)}"



import random
from base64 import b64decode, b64encode, urlsafe_b64decode
from collections.abc import AsyncGenerator, Coroutine, Generator, Iterable
from itertools import chain
from typing import Any, Literal, cast, overload
from uuid import uuid4

from google.api_core.exceptions import ResourceExhausted  # pyright: ignore[reportMissingImport]
from google.genai.types import (
    Candidate,
    ContentListUnionDict,
    FinishReason,
    FunctionCallingConfigMode,
    FunctionDeclarationDict,
    GenerateContentConfigDict,
    GenerateContentResponse,
    GenerateContentResponseUsageMetadata,
    HarmCategory,
    MediaResolution,
    Modality,
    Part,
    PartDict,
    SchemaDict,
    SpeechConfigDict,
)
from haiway import META_EMPTY, MISSING, ObservabilityLevel, as_dict, as_list, ctx

from draive.gemini.api import GeminiAPI
from draive.gemini.config import (
    GeminiConfig,
    GeminiSafetyConfig,
)
from draive.gemini.utils import unwrap_missing
from draive.models import (
    GenerativeModel,
    ModelContext,
    ModelContextElement,
    ModelInput,
    ModelInputBlocks,
    ModelInstructions,
    ModelOutput,
    ModelOutputBlock,
    ModelOutputBlocks,
    ModelOutputFailed,
    ModelOutputLimit,
    ModelOutputSelection,
    ModelRateLimit,
    ModelReasoning,
    ModelStreamOutput,
    ModelToolRequest,
    ModelToolResponse,
    ModelToolsDeclaration,
)
from draive.multimodal import ArtifactContent, MultimodalContent, TextContent
from draive.resources import ResourceContent, ResourceReference

__all__ = ("GeminiGenerating",)

# Consistent randomized backoff window for rate limits (seconds)
RATE_LIMIT_RETRY_RANGE: tuple[float, float] = (0.3, 3.0)


class GeminiGenerating(GeminiAPI):
    def generative_model(self) -> GenerativeModel:
        return GenerativeModel(generating=self.completion)

    @overload
    def completion(
        self,
        *,
        instructions: ModelInstructions,
        context: ModelContext,
        tools: ModelToolsDeclaration,
        output: ModelOutputSelection,
        stream: Literal[False] = False,
        config: GeminiConfig | None = None,
        **extra: Any,
    ) -> Coroutine[None, None, ModelOutput]: ...

    @overload
    def completion(
        self,
        *,
        instructions: ModelInstructions,
        context: ModelContext,
        tools: ModelToolsDeclaration,
        output: ModelOutputSelection,
        stream: Literal[True],
        config: GeminiConfig | None = None,
        **extra: Any,
    ) -> AsyncGenerator[ModelStreamOutput]: ...

    def completion(
        self,
        *,
        instructions: ModelInstructions,
        context: ModelContext,
        tools: ModelToolsDeclaration,
        output: ModelOutputSelection,
        stream: bool = False,
        config: GeminiConfig | None = None,
        **extra: Any,
    ) -> AsyncGenerator[ModelStreamOutput] | Coroutine[None, None, ModelOutput]:
        if stream:
            return self._completion_stream(
                instructions=instructions,
                context=context,
                tools=tools,
                output=output,
                config=config or ctx.state(GeminiConfig),
                **extra,
            )

        else:
            return self._completion(
                instructions=instructions,
                context=context,
                tools=tools,
                output=output,
                config=config or ctx.state(GeminiConfig),
                **extra,
            )

    async def _completion(
        self,
        *,
        instructions: ModelInstructions,
        context: ModelContext,
        tools: ModelToolsDeclaration,
        output: ModelOutputSelection,
        config: GeminiConfig,
        **extra: Any,
    ) -> ModelOutput:
        async with ctx.scope("model.completion"):
            ctx.record(
                ObservabilityLevel.INFO,
                attributes={
                    "model.provider": "gemini",
                    "model.name": config.model,
                    "model.temperature": config.temperature,
                    "model.output": str(output),
                    "model.max_output_tokens": config.max_output_tokens,
                    "model.thinking_budget": config.thinking_budget,
                    "model.tools.count": len(tools.specifications),
                    "model.tools.selection": tools.selection,
                    "model.stream": False,
                },
            )
            ctx.record(
                ObservabilityLevel.DEBUG,
                attributes={
                    "model.instructions": instructions,
                    "model.tools": [tool.name for tool in tools.specifications],
                    "model.context": [element.to_str() for element in context],
                },
            )

            request_config: GenerateContentConfigDict = _prepare_request_config(
                instructions=instructions,
                tools=tools,
                output=output,
                config=config,
            )
            request_content: ContentListUnionDict = [
                _context_element_as_content(element) for element in context
            ]

            try:
                completion: GenerateContentResponse = (
                    await self._client.aio.models.generate_content(
                        model=config.model,
                        config=request_config,
                        contents=request_content,
                    )
                )

            except ResourceExhausted as exc:
                ctx.record(
                    ObservabilityLevel.WARNING,
                    event="model.rate_limit",
                    attributes={
                        "model.provider": "gemini",
                        "model.name": config.model,
                    },
                )
                raise ModelRateLimit(
                    provider="gemini",
                    model=config.model,
                    retry_after=random.uniform(*RATE_LIMIT_RETRY_RANGE),  # nosec: B311
                ) from exc

            except Exception as exc:
                raise ModelOutputFailed(
                    provider="gemini",
                    model=config.model,
                    reason=str(exc),
                ) from exc

            _record_usage_metrics(
                completion.usage_metadata,
                model=config.model,
            )

            if not completion.candidates:
                raise ModelOutputFailed(
                    provider="gemini",
                    model=config.model,
                    reason="Missing candidates in response",
                )

            completion_candidate: Candidate = completion.candidates[0]  # we always request only one

            if completion_candidate.safety_ratings:
                ctx.record(
                    ObservabilityLevel.INFO,
                    event="model.safety.results",
                    attributes={
                        "results": [
                            f"{rating.category} |blocked: {rating.blocked}"
                            f" |probability:{rating.probability_score}"
                            f" |severity:{rating.severity_score}"
                            for rating in completion_candidate.safety_ratings
                            if rating.category
                        ],
                    },
                )

            if completion_candidate.finish_reason == FinishReason.SAFETY:
                raise ModelOutputFailed(
                    provider="gemini",
                    model=config.model,
                    reason=f"Safety filtering: {completion_candidate.finish_message}",
                )

            if completion_candidate.finish_reason not in (
                FinishReason.STOP,
                FinishReason.MAX_TOKENS,
            ):
                raise ModelOutputFailed(
                    provider="gemini",
                    model=config.model,
                    reason=f"Completion error: {completion_candidate.finish_message}",
                )

            if not completion_candidate.content or not completion_candidate.content.parts:
                raise ModelOutputFailed(
                    provider="gemini",
                    model=config.model,
                    reason="Empty completion content",
                )

            output_blocks: Iterable[ModelOutputBlock] = chain.from_iterable(
                _part_as_output_blocks(part) for part in completion_candidate.content.parts
            )

            if completion_candidate.finish_reason == FinishReason.MAX_TOKENS:
                # include the content collected so far and any tool requests
                raise ModelOutputLimit(
                    provider="gemini",
                    model=config.model,
                    max_output_tokens=unwrap_missing(config.max_output_tokens, default=0),
                    content=tuple(output_blocks),
                )

            return ModelOutput.of(
                *output_blocks,
                meta={
                    "identifier": completion.response_id,
                    "model": config.model,
                    "finish_reason": str(completion_candidate.finish_reason),
                },
            )

    async def _completion_stream(  # noqa: C901
        self,
        *,
        instructions: ModelInstructions,
        context: ModelContext,
        tools: ModelToolsDeclaration,
        output: ModelOutputSelection,
        config: GeminiConfig,
        **extra: Any,
    ) -> AsyncGenerator[ModelStreamOutput]:
        async with ctx.scope("model.completion.stream"):
            ctx.record(
                ObservabilityLevel.INFO,
                attributes={
                    "model.provider": "gemini",
                    "model.name": config.model,
                    "model.temperature": config.temperature,
                    "model.output": str(output),
                    "model.max_output_tokens": config.max_output_tokens,
                    "model.thinking_budget": config.thinking_budget,
                    "model.tools.count": len(tools.specifications),
                    "model.tools.selection": tools.selection,
                    "model.stream": True,
                },
            )
            ctx.record(
                ObservabilityLevel.DEBUG,
                attributes={
                    "model.instructions": instructions,
                    "model.tools": [tool.name for tool in tools.specifications],
                    "model.context": [element.to_str() for element in context],
                },
            )

            # Build request same as non-streaming
            request_config: GenerateContentConfigDict = _prepare_request_config(
                instructions=instructions,
                tools=tools,
                output=output,
                config=config,
            )
            request_content: ContentListUnionDict = [
                _context_element_as_content(element) for element in context
            ]

            try:
                response_stream = await self._client.aio.models.generate_content_stream(
                    model=config.model,
                    config=request_config,
                    contents=request_content,
                )

                collected_blocks: list[ModelOutputBlock] = []
                finish_reason: FinishReason | None = None
                finish_message: str | None = None

                async for chunk in response_stream:
                    _record_usage_metrics(
                        chunk.usage_metadata,
                        model=config.model,
                    )

                    if not chunk.candidates:
                        continue

                    chunk_candidate: Candidate = chunk.candidates[0]  # we always request only one
                    finish_reason = chunk_candidate.finish_reason
                    finish_message = chunk_candidate.finish_message

                    if chunk_candidate.safety_ratings:
                        ctx.record(
                            ObservabilityLevel.INFO,
                            event="model.safety.results",
                            attributes={
                                "results": [
                                    f"{rating.category} |blocked: {rating.blocked}"
                                    f" |probability:{rating.probability_score}"
                                    f" |severity:{rating.severity_score}"
                                    for rating in chunk_candidate.safety_ratings
                                    if rating.category
                                ],
                            },
                        )

                    if not chunk_candidate.content or not chunk_candidate.content.parts:
                        continue

                    for part in chunk_candidate.content.parts:
                        collected_blocks.extend(_part_as_output_blocks(part))
                        for element in _part_as_stream_elements(part):
                            # element is either a MultimodalContentElement or ModelToolRequest
                            yield element

                if finish_reason == FinishReason.SAFETY:
                    raise ModelOutputFailed(
                        provider="gemini",
                        model=config.model,
                        reason=f"Safety filtering: {finish_message or ''}",
                    )

                if finish_reason == FinishReason.MAX_TOKENS:
                    raise ModelOutputLimit(
                        provider="gemini",
                        model=config.model,
                        max_output_tokens=unwrap_missing(
                            config.max_output_tokens,
                            default=0,
                        ),
                        content=tuple(collected_blocks),
                    )

                if finish_reason not in (None, FinishReason.STOP):
                    raise ModelOutputFailed(
                        provider="gemini",
                        model=config.model,
                        reason=(
                            f"Completion error: {finish_message}"
                            if finish_message
                            else "Completion error"
                        ),
                    )

            except ResourceExhausted as exc:
                ctx.record(
                    ObservabilityLevel.WARNING,
                    event="model.rate_limit",
                    attributes={
                        "model.provider": "gemini",
                        "model.name": config.model,
                    },
                )
                # Propagate as ModelRateLimit with randomized backoff window
                raise ModelRateLimit(
                    provider="gemini",
                    model=config.model,
                    retry_after=random.uniform(*RATE_LIMIT_RETRY_RANGE),  # nosec: B311
                ) from exc

            except Exception as exc:
                # Convert to ModelOutputFailed for consistency
                raise ModelOutputFailed(
                    provider="gemini",
                    model=config.model,
                    reason=str(exc),
                ) from exc


def _record_usage_metrics(
    usage: GenerateContentResponseUsageMetadata | None,
    *,
    model: str,
) -> None:
    if usage is None:
        return

    ctx.record(
        ObservabilityLevel.INFO,
        metric="model.input_tokens",
        value=usage.prompt_token_count or 0,
        unit="tokens",
        kind="counter",
        attributes={
            "model.provider": "gemini",
            "model.name": model,
        },
    )
    ctx.record(
        ObservabilityLevel.INFO,
        metric="model.input_tokens.cached",
        value=usage.cached_content_token_count or 0,
        unit="tokens",
        kind="counter",
        attributes={
            "model.provider": "gemini",
            "model.name": model,
        },
    )
    ctx.record(
        ObservabilityLevel.INFO,
        metric="model.output_tokens",
        value=usage.candidates_token_count or 0,
        unit="tokens",
        kind="counter",
        attributes={
            "model.provider": "gemini",
            "model.name": model,
        },
    )


def _prepare_request_config(  # noqa: C901, PLR0912, PLR0915
    *,
    instructions: ModelInstructions,
    tools: ModelToolsDeclaration,
    output: ModelOutputSelection,
    config: GeminiConfig,
) -> GenerateContentConfigDict:
    temperature: float | None = unwrap_missing(config.temperature)
    max_output_tokens: int | None = unwrap_missing(config.max_output_tokens)
    seed: int | None = unwrap_missing(config.seed)

    response_schema: SchemaDict | None
    response_modalities: list[Modality] | None
    response_mime_type: str | None

    # Prefer explicit isinstance check for structured output to satisfy typing
    if isinstance(output, type):
        response_schema = cast(SchemaDict, output.__PARAMETERS_SPECIFICATION__)
        response_modalities = [Modality.TEXT]
        response_mime_type = "application/json"

    else:
        match output:
            case "auto":
                # not specified at all - use defaults
                response_schema = None
                response_modalities = None
                response_mime_type = None

            case "text":
                response_schema = None
                response_modalities = [Modality.TEXT]
                response_mime_type = "text/plain"

            case "json":
                response_schema = None
                response_modalities = [Modality.TEXT]
                response_mime_type = "application/json"

            case "image":
                response_schema = None
                response_modalities = [
                    Modality.TEXT,
                    Modality.IMAGE,
                ]  # google api does not allow to specify only image
                response_mime_type = None

            case "audio":
                response_schema = None
                response_modalities = [Modality.AUDIO]
                response_mime_type = None

            case "video":
                raise NotImplementedError("video output is not supported by Gemini")

            case ["text", "image"] | ["image", "text"]:
                response_schema = None
                response_modalities = [
                    Modality.TEXT,
                    Modality.IMAGE,
                ]
                response_mime_type = None

            case other:
                raise NotImplementedError(f"{other} output is not supported by Gemini")

    functions: list[FunctionDeclarationDict] | None
    function_calling_mode: FunctionCallingConfigMode | None
    allowed_function_names: list[str] | None = None
    if tools.specifications:
        functions = [
            FunctionDeclarationDict(
                name=tool.name,
                description=tool.description,
                parameters_json_schema=cast(SchemaDict, tool.parameters),
            )
            for tool in tools.specifications or []
        ]
        match tools.selection:
            case "auto":
                function_calling_mode = FunctionCallingConfigMode.AUTO

            case "required":
                function_calling_mode = FunctionCallingConfigMode.ANY

            case "none":
                functions = None  # no need to pass functions if none can be used
                function_calling_mode = FunctionCallingConfigMode.NONE

            case specific_tool:  # handle specific tool selection
                function_calling_mode = FunctionCallingConfigMode.ANY
                allowed_function_names = [specific_tool]

    else:  # no functions
        functions = None
        function_calling_mode = FunctionCallingConfigMode.NONE

    media_resolution: MediaResolution | None
    match config.media_resolution:
        case "low":
            media_resolution = MediaResolution.MEDIA_RESOLUTION_LOW

        case "medium":
            media_resolution = MediaResolution.MEDIA_RESOLUTION_MEDIUM

        case "high":
            media_resolution = MediaResolution.MEDIA_RESOLUTION_HIGH

        case _:
            media_resolution = None

    speech_config: SpeechConfigDict | None
    if config.speech_voice_name:
        speech_config = {
            "voice_config": {
                "prebuilt_voice_config": {
                    "voice_name": cast(str, config.speech_voice_name),
                },
            }
        }

    else:
        speech_config = None

    safety_config: GeminiSafetyConfig = unwrap_missing(
        config.safety,
        default=GeminiSafetyConfig(),
    )

    return {
        "temperature": temperature,
        "top_p": unwrap_missing(config.top_p),
        "top_k": unwrap_missing(config.top_k),
        "max_output_tokens": max_output_tokens,
        "candidate_count": 1,
        "seed": seed,
        "stop_sequences": unwrap_missing(
            config.stop_sequences,
            transform=as_list,
        ),
        "safety_settings": [
            {
                "category": HarmCategory.HARM_CATEGORY_HATE_SPEECH,
                "threshold": safety_config.harm_category_hate_speech_threshold,
            },
            {
                "category": HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT,
                "threshold": safety_config.harm_category_dangerous_content_threshold,
            },
            {
                "category": HarmCategory.HARM_CATEGORY_HARASSMENT,
                "threshold": safety_config.harm_category_harassment_threshold,
            },
            {
                "category": HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT,
                "threshold": safety_config.harm_category_sexually_explicit_threshold,
            },
            {
                "category": HarmCategory.HARM_CATEGORY_CIVIC_INTEGRITY,
                "threshold": safety_config.harm_category_civic_integrity_threshold,
            },
        ],
        "system_instruction": instructions,
        "tools": [{"function_declarations": functions}] if functions else None,
        "tool_config": {
            "function_calling_config": {
                "mode": function_calling_mode,
                **(
                    {"allowed_function_names": allowed_function_names}
                    if allowed_function_names
                    else {}
                ),
            }
        }
        if function_calling_mode
        else None,
        # prevent google from automatically resolving function calls
        "automatic_function_calling": {
            "disable": True,
            "maximum_remote_calls": None,
        }
        if function_calling_mode
        else None,
        "media_resolution": media_resolution,
        "response_modalities": response_modalities,
        "response_mime_type": response_mime_type,
        "response_schema": response_schema,
        "speech_config": speech_config,
        "thinking_config": {
            "include_thoughts": (cast(int, config.thinking_budget) > 0),
            "thinking_budget": config.thinking_budget,
        }
        if config.thinking_budget is not MISSING
        else None,
    }


def _context_element_as_content(
    element: ModelContextElement,
) -> dict[str, Any]:
    if isinstance(element, ModelInput):
        return {
            "role": "user",
            "parts": [*block_parts(element.blocks)],
        }

    else:
        assert isinstance(element, ModelOutput)  # nosec: B101
        return {
            "role": "model",
            "parts": [*block_parts(element.blocks)],
        }


def _part_as_output_blocks(
    part: Part,
) -> Generator[ModelOutputBlock]:
    if part.text:
        # assuming only text thinking is possible
        if part.thought:
            yield ModelReasoning.of(
                part.text,
                meta={
                    "kind": "thought",
                    "signature": b64encode(part.thought_signature).decode()
                    if part.thought_signature
                    else None,
                },
            )

        else:
            yield MultimodalContent.of(TextContent.of(part.text))

    if part.function_call and part.function_call.name:
        yield ModelToolRequest(
            identifier=part.function_call.id or uuid4().hex,
            tool=part.function_call.name,
            arguments=part.function_call.args if part.function_call.args is not None else {},
            meta=META_EMPTY,
        )

    if part.inline_data and part.inline_data.data:  # there is no content without content...
        yield MultimodalContent.of(
            ResourceContent.of(
                part.inline_data.data,
                mime_type=part.inline_data.mime_type or "application/octet-stream",
            )
        )

    if part.file_data and part.file_data.file_uri:  # there is no content without content...
        yield MultimodalContent.of(
            ResourceReference.of(
                part.file_data.file_uri,
                mime_type=part.file_data.mime_type,
            )
        )


def _part_as_stream_elements(
    part: Part,
) -> Generator[ModelStreamOutput]:
    if part.text:
        if part.thought:
            yield ModelReasoning.of(
                part.text,
                meta={
                    "kind": "thought",
                    "signature": b64encode(part.thought_signature).decode()
                    if part.thought_signature
                    else None,
                },
            )

        else:
            yield TextContent.of(part.text)

    if part.function_call and part.function_call.name:
        yield ModelToolRequest(
            identifier=part.function_call.id or uuid4().hex,
            tool=part.function_call.name,
            arguments=part.function_call.args if part.function_call.args is not None else {},
            meta=META_EMPTY,
        )

    if part.inline_data and part.inline_data.data:  # there is no content without content...
        yield ResourceContent.of(
            part.inline_data.data,
            mime_type=part.inline_data.mime_type or "application/octet-stream",
        )

    if part.file_data and part.file_data.file_uri:  # there is no content without content...
        yield ResourceReference.of(
            part.file_data.file_uri,
            mime_type=part.file_data.mime_type,
        )


def block_parts(
    blocks: ModelInputBlocks | ModelOutputBlocks,
    /,
) -> Generator[PartDict]:
    for block in blocks:
        if isinstance(block, ModelToolRequest):
            yield {
                "function_call": {
                    "id": block.identifier,
                    "name": block.tool,
                    "args": as_dict(block.arguments),
                }
            }

        elif isinstance(block, ModelToolResponse):
            yield {
                "function_response": {
                    "id": block.identifier,
                    "name": block.tool,
                    "response": {
                        "error": list(content_parts(block.content)),
                    }
                    if block.handling == "error"
                    else {
                        "output": list(content_parts(block.content)),
                    },
                }
            }

        elif isinstance(block, ModelReasoning):
            match block.meta.kind:
                case "thought":
                    if signature := block.meta.get_str("signature"):
                        yield {
                            "text": block.content.to_str(),
                            "thought": True,
                            "thought_signature": b64decode(signature),
                        }

                    else:
                        yield {
                            "text": block.content.to_str(),
                            "thought": True,
                        }

        else:
            yield from content_parts(block)


def content_parts(
    content: MultimodalContent,
    /,
) -> Generator[PartDict]:
    for part in content.parts:
        if isinstance(part, TextContent):
            yield {
                "text": part.text,
            }

        elif isinstance(part, ResourceContent):
            yield {
                "inline_data": {
                    # decode urlsafe base64 back to raw bytes for provider
                    "data": urlsafe_b64decode(part.data),
                    "mime_type": part.mime_type,
                },
            }

        elif isinstance(part, ResourceReference):
            yield {
                "file_data": {
                    "file_uri": part.uri,
                    "mime_type": part.mime_type,
                }
            }

        elif isinstance(part, ArtifactContent):
            # Skip artifacts that are marked as hidden
            if part.hidden:
                continue

            yield {
                "text": part.artifact.to_str(),
            }

        else:
            yield {
                "text": part.to_str(),
            }



# exporting haiway symbols for easier usage
from haiway import (
    META_EMPTY,
    MISSING,
    AsyncQueue,
    AsyncStream,
    AttributePath,
    AttributeRequirement,
    BasicValue,
    Configuration,
    ConfigurationInvalid,
    ConfigurationMissing,
    ConfigurationRepository,
    ContextPreset,
    Default,
    DefaultValue,
    Disposable,
    Disposables,
    EventSubscription,
    File,
    FileAccess,
    HTTPClient,
    HTTPClientError,
    HTTPHeaders,
    HTTPQueryParams,
    HTTPResponse,
    HTTPStatusCode,
    Immutable,
    LoggerObservability,
    Map,
    Meta,
    MetaTags,
    MetaValue,
    MetaValues,
    Missing,
    MissingContext,
    MissingState,
    Observability,
    ObservabilityAttribute,
    ObservabilityContext,
    ObservabilityLevel,
    ScopeContext,
    ScopeIdentifier,
    State,
    StateContext,
    ValidationError,
    always,
    as_dict,
    as_list,
    as_set,
    as_tuple,
    async_always,
    async_noop,
    asynchronous,
    cache,
    concurrently,
    ctx,
    execute_concurrently,
    getenv_base64,
    getenv_bool,
    getenv_float,
    getenv_int,
    getenv_str,
    is_missing,
    load_env,
    noop,
    not_missing,
    process_concurrently,
    retry,
    setup_logging,
    statemethod,
    stream_concurrently,
    throttle,
    timeout,
    unwrap_missing,
    without_missing,
)

from draive.conversation import (
    Conversation,
    ConversationEvent,
    ConversationInputChunk,
    ConversationMessage,
    ConversationOutputChunk,
    RealtimeConversation,
    RealtimeConversationSession,
)
from draive.embedding import (
    Embedded,
    ImageEmbedding,
    TextEmbedding,
    ValueEmbedding,
    mmr_vector_similarity_search,
    vector_similarity_score,
    vector_similarity_search,
)
from draive.generation import (
    AudioGeneration,
    ImageGeneration,
    ModelGeneration,
    TextGeneration,
)
from draive.guardrails import (
    GuardrailsAnonymization,
    GuardrailsAnonymizedContent,
    GuardrailsInputModerationException,
    GuardrailsModeration,
    GuardrailsModerationException,
    GuardrailsOutputModerationException,
    GuardrailsQualityException,
    GuardrailsQualityVerification,
)
from draive.models import (
    FunctionTool,
    GenerativeModel,
    Instructions,
    InstructionsArgumentDeclaration,
    InstructionsDeclaration,
    InstructionsDefining,
    InstructionsListing,
    InstructionsLoading,
    InstructionsMissing,
    InstructionsRemoving,
    InstructionsRepository,
    InstructionsTemplate,
    ModelContext,
    ModelContextElement,
    ModelException,
    ModelInput,
    ModelInputBlock,
    ModelInputBlocks,
    ModelInputChunk,
    ModelInputInvalid,
    ModelInstructions,
    ModelMemory,
    ModelMemoryRecall,
    ModelOutput,
    ModelOutputBlock,
    ModelOutputBlocks,
    ModelOutputChunk,
    ModelOutputFailed,
    ModelOutputInvalid,
    ModelOutputLimit,
    ModelOutputSelection,
    ModelRateLimit,
    ModelReasoning,
    ModelSession,
    ModelSessionEvent,
    ModelSessionInput,
    ModelSessionOutput,
    ModelSessionOutputSelection,
    ModelSessionScope,
    ModelStreamOutput,
    ModelToolFunctionSpecification,
    ModelToolHandling,
    ModelToolRequest,
    ModelToolResponse,
    ModelToolsDeclaration,
    ModelToolSpecification,
    ModelToolsSelection,
    RealtimeGenerativeModel,
    ResolveableInstructions,
    Tool,
    ToolAvailabilityChecking,
    Toolbox,
    ToolError,
    ToolErrorFormatting,
    ToolResultFormatting,
    ToolsLoading,
    ToolsProvider,
    ToolsSuggesting,
    instructions,
    tool,
)
from draive.multimodal import (
    ArtifactContent,
    Multimodal,
    MultimodalContent,
    MultimodalContentPart,
    MultimodalTag,
    TextContent,
)
from draive.parameters import (
    Argument,
    DataModel,
    Field,
    ParameterVerification,
)
from draive.resources import (
    MimeType,
    Resource,
    ResourceAvailabilityCheck,
    ResourceContent,
    ResourceCorrupted,
    ResourceDeleting,
    ResourceFetching,
    ResourceListFetching,
    ResourceMissing,
    ResourceReference,
    ResourceReferenceTemplate,
    ResourcesRepository,
    ResourceTemplate,
    ResourceUploading,
    resource,
)
from draive.splitters import split_text
from draive.stages import (
    Stage,
    StageException,
    StageState,
    stage,
)
from draive.utils import (
    Memory,
    VectorIndex,
)

__all__ = (
    "META_EMPTY",
    "MISSING",
    "Argument",
    "ArtifactContent",
    "AsyncQueue",
    "AsyncStream",
    "AttributePath",
    "AttributeRequirement",
    "AudioGeneration",
    "BasicValue",
    "Configuration",
    "ConfigurationInvalid",
    "ConfigurationMissing",
    "ConfigurationRepository",
    "ContextPreset",
    "Conversation",
    "ConversationEvent",
    "ConversationInputChunk",
    "ConversationMessage",
    "ConversationOutputChunk",
    "DataModel",
    "Default",
    "DefaultValue",
    "Disposable",
    "Disposables",
    "Embedded",
    "EventSubscription",
    "Field",
    "File",
    "FileAccess",
    "FunctionTool",
    "GenerativeModel",
    "GuardrailsAnonymization",
    "GuardrailsAnonymizedContent",
    "GuardrailsInputModerationException",
    "GuardrailsModeration",
    "GuardrailsModerationException",
    "GuardrailsOutputModerationException",
    "GuardrailsQualityException",
    "GuardrailsQualityVerification",
    "HTTPClient",
    "HTTPClientError",
    "HTTPHeaders",
    "HTTPQueryParams",
    "HTTPResponse",
    "HTTPStatusCode",
    "ImageEmbedding",
    "ImageGeneration",
    "Immutable",
    "Instructions",
    "InstructionsArgumentDeclaration",
    "InstructionsDeclaration",
    "InstructionsDefining",
    "InstructionsListing",
    "InstructionsLoading",
    "InstructionsMissing",
    "InstructionsRemoving",
    "InstructionsRepository",
    "InstructionsTemplate",
    "LoggerObservability",
    "Map",
    "Memory",
    "Meta",
    "MetaTags",
    "MetaValue",
    "MetaValues",
    "MimeType",
    "Missing",
    "MissingContext",
    "MissingState",
    "ModelContext",
    "ModelContextElement",
    "ModelException",
    "ModelGeneration",
    "ModelInput",
    "ModelInputBlock",
    "ModelInputBlocks",
    "ModelInputChunk",
    "ModelInputInvalid",
    "ModelInstructions",
    "ModelMemory",
    "ModelMemoryRecall",
    "ModelOutput",
    "ModelOutputBlock",
    "ModelOutputBlocks",
    "ModelOutputChunk",
    "ModelOutputFailed",
    "ModelOutputInvalid",
    "ModelOutputLimit",
    "ModelOutputSelection",
    "ModelRateLimit",
    "ModelReasoning",
    "ModelSession",
    "ModelSessionEvent",
    "ModelSessionInput",
    "ModelSessionOutput",
    "ModelSessionOutputSelection",
    "ModelSessionScope",
    "ModelStreamOutput",
    "ModelToolFunctionSpecification",
    "ModelToolHandling",
    "ModelToolRequest",
    "ModelToolResponse",
    "ModelToolSpecification",
    "ModelToolsDeclaration",
    "ModelToolsSelection",
    "Multimodal",
    "MultimodalContent",
    "MultimodalContentPart",
    "MultimodalTag",
    "Observability",
    "ObservabilityAttribute",
    "ObservabilityContext",
    "ObservabilityLevel",
    "ParameterVerification",
    "RealtimeConversation",
    "RealtimeConversationSession",
    "RealtimeGenerativeModel",
    "ResolveableInstructions",
    "Resource",
    "Resource",
    "ResourceAvailabilityCheck",
    "ResourceContent",
    "ResourceContent",
    "ResourceCorrupted",
    "ResourceDeleting",
    "ResourceFetching",
    "ResourceListFetching",
    "ResourceMissing",
    "ResourceMissing",
    "ResourceReference",
    "ResourceReference",
    "ResourceReferenceTemplate",
    "ResourceTemplate",
    "ResourceTemplate",
    "ResourceUploading",
    "ResourcesRepository",
    "ScopeContext",
    "ScopeIdentifier",
    "Stage",
    "StageException",
    "StageState",
    "State",
    "StateContext",
    "TextContent",
    "TextEmbedding",
    "TextGeneration",
    "Tool",
    "ToolAvailabilityChecking",
    "ToolError",
    "ToolErrorFormatting",
    "ToolResultFormatting",
    "Toolbox",
    "ToolsLoading",
    "ToolsProvider",
    "ToolsSuggesting",
    "ValidationError",
    "ValueEmbedding",
    "VectorIndex",
    "always",
    "as_dict",
    "as_list",
    "as_set",
    "as_tuple",
    "async_always",
    "async_noop",
    "asynchronous",
    "cache",
    "concurrently",
    "ctx",
    "execute_concurrently",
    "getenv_base64",
    "getenv_bool",
    "getenv_float",
    "getenv_int",
    "getenv_str",
    "instructions",
    "is_missing",
    "load_env",
    "mmr_vector_similarity_search",
    "noop",
    "not_missing",
    "process_concurrently",
    "resource",
    "resource",
    "retry",
    "setup_logging",
    "split_text",
    "stage",
    "statemethod",
    "stream_concurrently",
    "throttle",
    "timeout",
    "tool",
    "unwrap_missing",
    "vector_similarity_score",
    "vector_similarity_search",
    "without_missing",
)



from draive.models.tools.function import FunctionTool, tool
from draive.models.tools.provider import ToolsProvider
from draive.models.tools.toolbox import Toolbox
from draive.models.tools.types import (
    ModelToolSpecification,
    Tool,
    ToolAvailabilityChecking,
    ToolError,
    ToolErrorFormatting,
    ToolResultFormatting,
    ToolsLoading,
    ToolsSuggesting,
)

__all__ = (
    "FunctionTool",
    "ModelToolSpecification",
    "Tool",
    "ToolAvailabilityChecking",
    "ToolError",
    "ToolErrorFormatting",
    "ToolResultFormatting",
    "Toolbox",
    "ToolsLoading",
    "ToolsProvider",
    "ToolsSuggesting",
    "tool",
)



from collections.abc import Sequence
from typing import Any, cast

from haiway import Meta, cache, ctx
from haiway.postgres import Postgres, PostgresRow

from draive.models.instructions import (
    InstructionsDeclaration,
    InstructionsRepository,
)
from draive.models.instructions.types import InstructionsArgumentDeclaration

__all__ = ("PostgresInstructionsRepository",)


# POSTGRES SCHEMA
#
# CREATE TABLE instructions (
#     name TEXT NOT NULL,
#     description TEXT DEFAULT NULL,
#     content TEXT NOT NULL,
#     arguments JSONB NOT NULL DEFAULT '[]'::jsonb,
#     meta JSONB NOT NULL DEFAULT '{}'::jsonb,
#     created TIMESTAMPTZ NOT NULL DEFAULT CURRENT_TIMESTAMP,
#     PRIMARY KEY (name, created)
# );
#


def PostgresInstructionsRepository(
    cache_limit: int = 32,
    cache_expiration: float = 600.0,  # 10 min
) -> InstructionsRepository:
    """Return a Postgres-backed instructions repository with caching.

    Parameters
    ----------
    cache_limit
        Maximum number of loaded instruction payloads cached concurrently.
    cache_expiration
        Lifetime in seconds for cached entries before reloading from Postgres.

    Returns
    -------
    InstructionsRepository
        Repository facade operating on the ``instructions`` Postgres table.
    """

    @cache(
        limit=1,
        expiration=cache_expiration,
    )
    async def listing(
        **extra: Any,
    ) -> Sequence[InstructionsDeclaration]:
        ctx.log_info("Listing instructions...")
        results: Sequence[PostgresRow] = await Postgres.fetch(
            """
            SELECT DISTINCT ON (name)
                name::TEXT,
                description::TEXT,
                arguments::TEXT,
                meta::TEXT

            FROM
                instructions

            ORDER BY
                name,
                created
            DESC;
            """
        )
        ctx.log_info(f"...{len(results)} results found!")

        return tuple(
            InstructionsDeclaration(
                name=cast(str, result["name"]),
                description=cast(str, result["description"] or ""),
                arguments=InstructionsArgumentDeclaration.from_json_array(
                    cast(str, result["arguments"] or "[]")
                ),
                meta=Meta.from_json(cast(str, result["meta"] or "{}")),
            )
            for result in results
        )

    @cache(
        limit=cache_limit,
        expiration=cache_expiration,
    )
    async def load(
        name: str,
        /,
    ) -> str | None:
        ctx.log_info(f"Loading '{name}' instructions ...")
        result = await Postgres.fetch_one(
            """
            SELECT DISTINCT ON (name)
                content

            FROM
                instructions

            WHERE
                name = $1

            ORDER BY
                name,
                created
            DESC

            LIMIT 1;
            """,
            name,
        )

        if not result:
            ctx.log_info("...instructions not found!")
            return None

        ctx.log_info("...instructions loaded!")
        return cast(str, result["content"])

    async def loading(
        name: str,
        meta: Meta,
        **extra: Any,
    ) -> str | None:
        return await load(name)

    async def defining(
        declaration: InstructionsDeclaration,
        content: str,
        **extra: Any,
    ) -> None:
        ctx.log_info(f"Defining '{declaration.name}' instruction...")
        await Postgres.execute(
            """
            INSERT INTO
                instructions (
                    name,
                    description,
                    content,
                    arguments,
                    meta
                )

            VALUES
                (
                    $1::TEXT,
                    $2::TEXT,
                    $3::TEXT,
                    $4::JSONB,
                    $5::JSONB
                );
            """,
            declaration.name,
            declaration.description or None,
            content,
            f"[{','.join(argument.to_json() for argument in declaration.arguments)}]",
            declaration.meta.to_json(),
        )
        ctx.log_info("...clearing cache...")
        await load.clear_cache()
        await listing.clear_cache()
        ctx.log_info("...instructions definition completed!")

    async def removing(
        name: str,
        meta: Meta,
        **extra: Any,
    ) -> None:
        ctx.log_info(f"Removing '{name}' instructions...")
        await Postgres.execute(
            """
            DELETE FROM
                instructions

            WHERE
                name = $1;
            """,
            name,
        )

        ctx.log_info("...clearing cache...")
        await load.clear_cache()
        await listing.clear_cache()
        ctx.log_info("...instructions removal completed!")

    return InstructionsRepository(
        listing=listing,
        loading=loading,
        defining=defining,
        removing=removing,
    )



from typing import Self, final

from haiway import META_EMPTY, Meta, MetaValues

from draive.parameters import DataModel

__all__ = ("TextContent",)


@final
class TextContent(DataModel):
    """
    Text block with optional metadata.

    The ``TextContent`` model represents a unit of plain text in a
    ``MultimodalContent`` sequence. The text is kept immutable and can be
    associated with structured metadata to support filtering, grouping, and
    selection during downstream processing.

    Attributes
    ----------
    text : str
        The textual content.
    meta : Meta
        Structured metadata associated with the text. Used for filtering and
        grouping operations. Defaults to an empty metadata object.
    """

    @classmethod
    def of(
        cls,
        text: str,
        *,
        meta: Meta | MetaValues | None = None,
    ) -> Self:
        """
        Construct a ``TextContent`` instance.

        Parameters
        ----------
        text : str
            The textual content value.
        meta : Meta | MetaValues | None, optional
            Optional metadata to attach. Accepts an existing ``Meta`` instance,
            a mapping of metadata key-value pairs, or ``None`` to use empty
            metadata.

        Returns
        -------
        Self
            A new ``TextContent`` instance with normalized metadata.
        """
        return cls(
            text=text,
            meta=Meta.of(meta),
        )

    text: str
    meta: Meta = META_EMPTY

    def to_str(self) -> str:
        """
        Produce the textual representation of this element.

        Returns
        -------
        str
            The raw ``text`` value without metadata.
        """
        return self.text

    def __bool__(self) -> bool:
        """
        Indicate whether the text is non-empty.

        Returns
        -------
        bool
            ``True`` if ``text`` contains at least one character, ``False``
            otherwise. Whitespace-only strings are considered truthy.
        """
        return len(self.text) > 0



from base64 import urlsafe_b64decode
from typing import Any, Literal

from haiway import ObservabilityLevel, asynchronous, ctx

from draive.bedrock.api import BedrockAPI
from draive.bedrock.config import BedrockInputGuardraisConfig, BedrockOutputGuardraisConfig
from draive.guardrails import GuardrailsModeration, GuardrailsModerationException
from draive.multimodal import Multimodal, MultimodalContent, TextContent
from draive.resources import ResourceContent, ResourceReference

__all__ = ("BedrockGuardrails",)


class BedrockGuardrails(BedrockAPI):
    def guardrails_moderation(self) -> GuardrailsModeration:
        return GuardrailsModeration(
            input_checking=self.content_input_verification,
            output_checking=self.content_output_verification,
        )

    async def content_input_verification(
        self,
        content: Multimodal,
        /,
        *,
        config: BedrockInputGuardraisConfig | None = None,
        **extra: Any,
    ) -> None:
        guardrails_config: BedrockInputGuardraisConfig = config or ctx.state(
            BedrockInputGuardraisConfig
        )

        content = MultimodalContent.of(content)
        await self._content_input_verification(
            content,
            config=guardrails_config,
        )

    @asynchronous
    def _content_input_verification(
        self,
        content: MultimodalContent,
        /,
        *,
        config: BedrockInputGuardraisConfig,
        **extra: Any,
    ) -> None:
        response: dict[str, Any] = self._client.apply_guardrail(
            guardrailIdentifier=config.guardrail_identifier,
            guardrailVersion=config.guardrail_version,
            source="INPUT",
            content=self._convert_content(
                content,
                qualifier="query",
            ),
        )

        if response.get("action") == "GUARDRAIL_INTERVENED":
            raise GuardrailsModerationException(
                f"Content violated guardrails ({config.guardrail_identifier}) rules",
                violations={
                    filter["type"]: 1.0  # bedrock do not provide scores in response
                    for assessment in response.get("assessments", ())
                    for filter in assessment.get("contentPolicy", {}).get("filters", ())  # noqa: A001
                    if "type" in filter and filter.get("detected", False)
                },
                content=content,
            )

    async def content_output_verification(
        self,
        content: Multimodal,
        /,
        *,
        config: BedrockOutputGuardraisConfig | None = None,
        **extra: Any,
    ) -> None:
        guardrails_config: BedrockOutputGuardraisConfig = config or ctx.state(
            BedrockOutputGuardraisConfig
        )
        async with ctx.scope("bedrock_guardrails"):
            ctx.record(
                ObservabilityLevel.INFO,
                attributes={
                    "guardrails.provider": "bedrock",
                    "guardrails.identifier": guardrails_config.guardrail_identifier,
                    "guardrails.version": guardrails_config.guardrail_version,
                },
            )

            content = MultimodalContent.of(content)
            await self._content_output_verification(
                content,
                config=guardrails_config,
            )

    @asynchronous
    def _content_output_verification(
        self,
        content: MultimodalContent,
        /,
        *,
        config: BedrockOutputGuardraisConfig,
        **extra: Any,
    ) -> None:
        response: dict[str, Any] = self._client.apply_guardrail(
            guardrailIdentifier=config.guardrail_identifier,
            guardrailVersion=config.guardrail_version,
            source="OUTPUT",
            content=self._convert_content(
                content,
                qualifier="guard_content",
            ),
            outputScope="INTERVENTIONS",
        )

        if response.get("action") == "GUARDRAIL_INTERVENED":
            raise GuardrailsModerationException(
                f"Violated guardrails rules: {response.get('actionReason')}",
                # bedrock do not provide scores in response
                violations={config.guardrail_identifier: 1.0},
                content=content,
            )

    def _convert_content(
        self,
        content: MultimodalContent,
        /,
        *,
        qualifier: Literal["guard_content", "query"],
    ) -> list[dict[str, Any]]:
        moderated_content: list[dict[str, Any]] = []
        for part in content.parts:
            match part:
                case TextContent() as text:
                    moderated_content.append(
                        {
                            "text": {
                                "text": text.text,
                                "qualifiers": [qualifier],
                            },
                        }
                    )

                case ResourceContent() as content_part:
                    # Only selected image resources are supported by Bedrock guardrails
                    if content_part.mime_type == "image/jpeg":
                        moderated_content.append(
                            {
                                "image": {
                                    "format": "jpeg",
                                    "source": {"bytes": urlsafe_b64decode(content_part.data)},
                                }
                            }
                        )

                    elif content_part.mime_type == "image/png":
                        moderated_content.append(
                            {
                                "image": {
                                    "format": "png",
                                    "source": {"bytes": urlsafe_b64decode(content_part.data)},
                                }
                            }
                        )

                    elif content_part.mime_type == "image/gif":
                        moderated_content.append(
                            {
                                "image": {
                                    "format": "gif",
                                    "source": {"bytes": urlsafe_b64decode(content_part.data)},
                                }
                            }
                        )

                    else:
                        ctx.log_warning(
                            "Attempting to use guardrails on unsupported media "
                            f"({content_part.mime_type}), verifying as text..."
                        )
                        moderated_content.append(
                            {
                                "text": {
                                    "text": content_part.to_str(),
                                    "qualifiers": [qualifier],
                                }
                            }
                        )

                case ResourceReference() as ref:
                    ctx.log_warning(
                        "Bedrock guardrails require image bytes; got ResourceReference. "
                        "Verifying reference as JSON text."
                    )
                    moderated_content.append(
                        {
                            "text": {
                                "text": ref.to_str(),
                                "qualifiers": [qualifier],
                            }
                        }
                    )

                case other:
                    ctx.log_warning(
                        f"Attempting to use guardrails on unsupported content ({type(other)}),"
                        " verifying as text..."
                    )
                    moderated_content.append(
                        {
                            "text": {
                                "text": other.to_str(),
                                "qualifiers": [qualifier],
                            }
                        }
                    )

        return moderated_content



from collections.abc import Collection, Iterable
from types import TracebackType
from typing import Any, final

from haiway import State

from draive.embedding import TextEmbedding
from draive.generation import ImageGeneration
from draive.guardrails import GuardrailsModeration
from draive.models import GenerativeModel, RealtimeGenerativeModel
from draive.openai.api import OpenAIAPI
from draive.openai.embedding import OpenAIEmbedding
from draive.openai.images import OpenAIImageGeneration
from draive.openai.moderation import OpenAIContentModeration
from draive.openai.realtime import OpenAIRealtime
from draive.openai.responses import OpenAIResponses

__all__ = ("OpenAI",)


@final
class OpenAI(
    OpenAIResponses,
    OpenAIRealtime,
    OpenAIEmbedding,
    OpenAIImageGeneration,
    OpenAIContentModeration,
    OpenAIAPI,
):
    """
    Access to OpenAI services, can be used to prepare various functionalities like lmm.
    """

    __slots__ = ("_features",)

    def __init__(
        self,
        base_url: str | None = None,
        api_key: str | None = None,
        organization: str | None = None,
        azure_api_endpoint: str | None = None,
        azure_api_version: str | None = None,
        azure_deployment: str | None = None,
        timeout: float | None = None,
        features: Collection[type[State]] | None = None,
        **extra: Any,
    ) -> None:
        super().__init__(
            base_url=base_url,
            api_key=api_key,
            organization=organization,
            azure_api_endpoint=azure_api_endpoint,
            azure_api_version=azure_api_version,
            azure_deployment=azure_deployment,
            timeout=timeout,
            **extra,
        )

        self._features: frozenset[type[State]] = (
            frozenset(features)
            if features is not None
            else frozenset(
                (
                    GenerativeModel,
                    RealtimeGenerativeModel,
                    TextEmbedding,
                    ImageGeneration,
                    GuardrailsModeration,
                )
            )
        )

    async def __aenter__(self) -> Iterable[State]:
        await self._initialize_client()
        state: list[State] = []
        if GenerativeModel in self._features:
            state.append(self.generative_model())

        if RealtimeGenerativeModel in self._features:
            state.append(self.realtime_generative_model())

        if TextEmbedding in self._features:
            state.append(self.text_embedding())

        if ImageGeneration in self._features:
            state.append(self.image_generation())

        if GuardrailsModeration in self._features:
            state.append(self.moderation_guardrails())

        return state

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._deinitialize_client()



from collections.abc import Collection, Iterable
from types import TracebackType
from typing import final

from haiway import State

from draive.bedrock.api import BedrockAPI
from draive.bedrock.converse import BedrockConverse
from draive.bedrock.guardrails import BedrockGuardrails
from draive.guardrails import GuardrailsModeration
from draive.models import GenerativeModel

__all__ = ("Bedrock",)


@final
class Bedrock(
    BedrockConverse,
    BedrockGuardrails,
    BedrockAPI,
):
    __slots__ = ("_features",)

    def __init__(
        self,
        *,
        features: Collection[type[GenerativeModel | GuardrailsModeration]] | None = None,
        aws_region: str | None = None,
    ) -> None:
        super().__init__(aws_region=aws_region)
        self._features: frozenset[type[State]] = (
            frozenset(features)
            if features is not None
            else frozenset((GenerativeModel, GuardrailsModeration))
        )

    async def __aenter__(self) -> Iterable[State]:
        await self._initialize_client()
        state: list[State] = []
        if GenerativeModel in self._features:
            state.append(self.generative_model())

        if GuardrailsModeration in self._features:
            state.append(self.guardrails_moderation())

        return state

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self._deinitialize_client()



from asyncio import gather
from base64 import urlsafe_b64decode
from collections.abc import AsyncGenerator, Callable, Collection, Coroutine, Mapping, Sequence
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from itertools import chain
from types import TracebackType
from typing import Any, Self, cast, final
from urllib.parse import ParseResult, urlparse, urlunparse
from uuid import uuid4

from haiway import BasicValue, Meta, MetaTags, as_dict, as_list, ctx
from mcp import ClientSession, ListToolsResult, StdioServerParameters, stdio_client
from mcp import Tool as MCPTool
from mcp.client.sse import sse_client
from mcp.types import AudioContent as MCPAudioContent
from mcp.types import (
    BlobResourceContents,
    CallToolResult,
    ListResourcesResult,
    ReadResourceResult,
    TextResourceContents,
)
from mcp.types import EmbeddedResource as MCPEmbeddedResource
from mcp.types import ImageContent as MCPImageContent
from mcp.types import ResourceLink as MCPResourceLink
from mcp.types import TextContent as MCPTextContent
from pydantic import AnyUrl

from draive.models import FunctionTool, ModelToolSpecification, Tool, ToolError, ToolsProvider
from draive.multimodal import ArtifactContent, MultimodalContent, TextContent
from draive.parameters import DataModel, ToolParametersSpecification
from draive.resources import ResourceContent, ResourceReference, ResourcesRepository

__all__ = (
    "MCPClient",
    "MCPClients",
)


@final
class MCPClient:
    @classmethod
    def stdio(
        cls,
        identifier: str | None = None,
        *,
        command: str,
        args: Sequence[str] | None = None,
        env: Mapping[str, str] | None = None,
        features: Collection[type[ResourcesRepository] | type[ToolsProvider]] | None = None,
        tags: MetaTags | None = None,
    ) -> Self:
        @asynccontextmanager
        async def mcp_stdio_session() -> AsyncGenerator[ClientSession]:
            async with stdio_client(
                StdioServerParameters(
                    command=command,
                    args=as_list(args) if args is not None else [],
                    env=as_dict(env),
                )
            ) as (read, write):
                async with ClientSession(read, write) as session:
                    yield session

        return cls(
            identifier or uuid4().hex,
            session_manager=mcp_stdio_session(),
            features=features if features is not None else (ResourcesRepository, ToolsProvider),
            tags=tags if tags is not None else (),
        )

    @classmethod
    def sse(
        cls,
        identifier: str | None = None,
        *,
        url: str,
        headers: Mapping[str, Any] | None = None,
        timeout: float = 5,
        sse_read_timeout: float = 60 * 5,
        features: Collection[type[ResourcesRepository] | type[ToolsProvider]] | None = None,
        tags: MetaTags | None = None,
    ) -> Self:
        @asynccontextmanager
        async def mcp_sse_session() -> AsyncGenerator[ClientSession]:
            async with sse_client(
                url=url,
                headers=as_dict(headers),
                timeout=timeout,
                sse_read_timeout=sse_read_timeout,
            ) as (read, write):
                async with ClientSession(read, write) as session:
                    yield session

        return cls(
            identifier or uuid4().hex,
            session_manager=mcp_sse_session(),
            features=features if features is not None else (ResourcesRepository, ToolsProvider),
            tags=tags if tags is not None else (),
        )

    __slots__ = (
        "_features",
        "_session",
        "_session_manager",
        "identifier",
        "tags",
    )

    def __init__(
        self,
        identifier: str,
        *,
        session_manager: AbstractAsyncContextManager[ClientSession],
        features: Collection[type[ResourcesRepository] | type[ToolsProvider]],
        tags: MetaTags,
    ) -> None:
        self.identifier: str = identifier
        self._session_manager: AbstractAsyncContextManager[ClientSession] = session_manager
        self._session: ClientSession
        self._features: Collection[type[ResourcesRepository] | type[ToolsProvider]] = features
        self.tags: MetaTags = tags

    async def resources_list(
        self,
        **extra: Any,
    ) -> Sequence[ResourceReference]:
        assert hasattr(  # nosec: B101
            self,
            "_session",
        ), "MCPClient has to be initialized through async context entering"

        # TODO: in theory there are some extra elements like pagination
        # we are not supporting it as for now (how to pass cursor anyways?)
        result: ListResourcesResult = await self._session.list_resources()

        return [
            ResourceReference(
                uri=self._with_uri_identifier(resource.uri.unicode_string()),
                mime_type=resource.mimeType,
                meta=Meta(
                    {
                        "mcp_server": self.identifier,
                        "tags": self.tags,
                        "name": resource.name,
                        "description": resource.description,
                    }
                ),
            )
            for resource in result.resources
        ]

    async def resource_fetch(
        self,
        uri: str,
        **extra: Any,
    ) -> Collection[ResourceReference] | ResourceContent | None:
        assert hasattr(  # nosec: B101
            self,
            "_session",
        ), "MCPClient has to be initialized through async context entering"

        result: ReadResourceResult = await self._session.read_resource(
            uri=AnyUrl(self._without_uri_identifier(uri))
        )

        match result.contents:
            case [resource]:
                # if there is only a single element return it directly
                return self._convert_resource_content(resource)

            case [*resources]:
                # otherwise convert to references ignoring content
                return [
                    ResourceReference(
                        uri=resource.uri.unicode_string(),
                        mime_type=resource.mimeType,
                        meta=Meta(
                            {
                                "mcp_server": self.identifier,
                                "tags": self.tags,
                            }
                        ),
                    )
                    for resource in resources
                ]

    async def resource_upload(
        self,
        uri: str,
        content: ResourceContent,
        **extra: Any,
    ) -> Meta:
        raise NotImplementedError("Resource uploading is not supported by MCP servers")

    async def resource_delete(
        self,
        uri: str,
        **extra: Any,
    ) -> None:
        raise NotImplementedError("Resource deleting is not supported by MCP servers")

    async def tools_fetch(
        self,
        **extra: Any,
    ) -> Sequence[Tool]:
        tools: ListToolsResult = await self._session.list_tools()

        return tuple(
            _convert_tool(
                tool,
                tool_call=self._tool_call,
                source=self.identifier,
                tags=self.tags,
            )
            for tool in tools.tools
        )

    async def _tool_call(
        self,
        name: str,
        arguments: Mapping[str, BasicValue],
    ) -> MultimodalContent:
        result: CallToolResult = await self._session.call_tool(
            name=name,
            arguments=as_dict(arguments),
        )
        content: MultimodalContent = MultimodalContent.of(
            *await gather(*(_convert_content(part) for part in result.content))
        )

        if result.isError:
            raise ToolError(
                f"Remote tool {name} failed",
                content=content,
            )

        return content

    def _with_uri_identifier(
        self,
        uri: str,
        /,
    ) -> str:
        if not uri:
            return uri

        parsed: ParseResult = urlparse(uri)
        if parsed.netloc:
            return urlunparse(
                (
                    parsed.scheme,
                    f"{self.identifier}.{parsed.netloc}",
                    parsed.path,
                    parsed.params,
                    parsed.query,
                    parsed.fragment,
                )
            )
        else:
            # Ensure path starts with /
            path: str = parsed.path
            if not path.startswith("/"):
                path = "/" + path

            return urlunparse(
                (
                    # default to mcp scheme if empty
                    parsed.scheme or "mcp",
                    # use identifier as domain/netloc
                    self.identifier,
                    path,
                    parsed.params,
                    parsed.query,
                    parsed.fragment,
                )
            )

    def _without_uri_identifier(
        self,
        uri: str,
        /,
    ) -> str:
        if not uri:
            return uri

        parsed: ParseResult = urlparse(uri)
        if not parsed.netloc:
            return uri

        if parsed.netloc == self.identifier:
            return urlunparse(
                (
                    parsed.scheme if parsed.scheme != "mcp" else "",
                    "",
                    parsed.path,
                    parsed.params,
                    parsed.query,
                    parsed.fragment,
                )
            )

        match parsed.netloc.split(".", 1):
            case [identifier, netloc] if identifier == self.identifier:
                return urlunparse(
                    (
                        parsed.scheme,
                        netloc,
                        parsed.path,
                        parsed.params,
                        parsed.query,
                        parsed.fragment,
                    )
                )

            case _:
                return uri

    def _convert_resource_content(
        self,
        resource: TextResourceContents | BlobResourceContents,
        /,
    ) -> ResourceContent:
        match resource:
            case TextResourceContents() as text_resource:
                return ResourceContent(
                    data=urlsafe_b64decode(text_resource.text.encode()).decode(),
                    mime_type=text_resource.mimeType or "text/plain",
                    meta=Meta(
                        {
                            "mcp_server": self.identifier,
                            "tags": self.tags,
                        }
                    ),
                )

            case BlobResourceContents() as blob_resource:
                return ResourceContent(
                    data=blob_resource.blob,
                    mime_type=blob_resource.mimeType or "application/octet-stream",
                    meta=Meta(
                        {
                            "mcp_server": self.identifier,
                            "tags": self.tags,
                        }
                    ),
                )

    async def __aenter__(self) -> Sequence[ResourcesRepository | ToolsProvider]:
        self._session = await self._session_manager.__aenter__()
        await self._session.initialize()

        features: list[ResourcesRepository | ToolsProvider] = []
        if ResourcesRepository in self._features:
            features.append(
                ResourcesRepository(
                    list_fetching=self.resources_list,
                    fetching=self.resource_fetch,
                    uploading=self.resource_upload,
                    deleting=self.resource_delete,
                    meta=Meta(
                        {
                            "mcp_server": self.identifier,
                            "tags": self.tags,
                        }
                    ),
                )
            )

        if ToolsProvider in self._features:
            features.append(
                ToolsProvider(
                    loading=self.tools_fetch,
                    meta=Meta(
                        {
                            "mcp_server": self.identifier,
                            "tags": self.tags,
                        }
                    ),
                )
            )

        return features

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._session_manager.__aexit__(
            exc_type,
            exc_val,
            exc_tb,
        )
        del self._session


async def _convert_content(  # noqa: C901, PLR0911
    content: MCPTextContent
    | MCPImageContent
    | MCPAudioContent
    | MCPResourceLink
    | MCPEmbeddedResource,
    /,
) -> MultimodalContent:
    match content:
        case MCPTextContent() as text:
            return MultimodalContent.of(text.text)

        case MCPImageContent() as image:
            return MultimodalContent.of(
                ResourceContent.of(
                    urlsafe_b64decode(image.data),
                    mime_type=image.mimeType,
                )
            )

        case MCPAudioContent() as audio:
            return MultimodalContent.of(
                ResourceContent.of(
                    urlsafe_b64decode(audio.data),
                    mime_type=audio.mimeType,
                )
            )

        case MCPEmbeddedResource() as resource:
            match resource.resource:
                case TextResourceContents() as text:
                    return MultimodalContent.of(TextContent(text=text.text))

                case BlobResourceContents() as blob:
                    match blob.mimeType:
                        case None:
                            raise NotImplementedError(
                                "Unsupported embedded resource - missing mime!"
                            )

                        case "text/plain":
                            return MultimodalContent.of(
                                TextContent(text=urlsafe_b64decode(blob.blob).decode())
                            )

                        case "application/json":
                            return MultimodalContent.of(
                                ArtifactContent.of(
                                    DataModel.from_json(urlsafe_b64decode(blob.blob).decode())
                                )
                            )

                        case other:
                            # try to match supported media or raise an exception
                            return MultimodalContent.of(
                                ResourceContent.of(
                                    urlsafe_b64decode(blob.blob),
                                    mime_type=other,
                                )
                            )

        case MCPResourceLink():
            raise NotImplementedError("MCP resource links are not supported yet")


def _convert_tool(
    mcp_tool: MCPTool,
    /,
    *,
    tool_call: Callable[[str, Mapping[str, BasicValue]], Coroutine[None, None, MultimodalContent]],
    source: str,
    tags: MetaTags,
) -> Tool:
    name: str = mcp_tool.name

    async def remote_call(**arguments: Any) -> MultimodalContent:
        return await tool_call(
            name,
            arguments,
        )

    return FunctionTool(
        name=name,
        description=mcp_tool.description,
        parameters=cast(
            ToolParametersSpecification,
            {
                **mcp_tool.inputSchema,
                "additionalProperties": False,
            },
        ),
        function=remote_call,
        availability=_available,
        result_formatting=_format_tool_result,
        error_formatting=_format_tool_failure,
        handling="response",
        meta=Meta(
            {
                "mcp_server": source,
                "tags": tags,
            }
        ),
    )


def _available(
    tools_turn: int,
    specification: ModelToolSpecification,
) -> bool:
    return True


def _format_tool_result(
    result: Any,
) -> MultimodalContent:
    assert isinstance(result, MultimodalContent)  # nosec: B101
    return result


def _format_tool_failure(
    error: Exception,
) -> MultimodalContent:
    if isinstance(error, ToolError):
        return error.content

    return MultimodalContent.of("ERROR")


@final
class MCPClients:
    __slots__ = (
        "_clients",
        "_prompts",
        "_resources",
        "_tools",
    )

    def __init__(
        self,
        client: MCPClient,
        *clients: MCPClient,
    ) -> None:
        self._clients: Mapping[str, MCPClient] = {c.identifier: c for c in [client, *clients]}
        self._resources: Mapping[str, ResourcesRepository]
        self._tools: Mapping[str, ToolsProvider]

    async def resources_list(
        self,
        *,
        mcp_server: str | None = None,
        **extra: Any,
    ) -> Sequence[ResourceReference]:
        if mcp_server is None:
            return tuple(
                chain.from_iterable(
                    await gather(
                        *[client.fetch_list(**extra) for client in self._resources.values()]
                    )
                )
            )

        elif resources := self._resources.get(mcp_server):
            return await resources.fetch_list(**extra)

        else:
            return ()

    async def resource_fetch(
        self,
        uri: str,
        **extra: Any,
    ) -> Collection[ResourceReference] | ResourceContent | None:
        if client_identifier := self._client_identifier_for_uri(uri):
            return await self._resources[client_identifier].fetching(uri)

        else:
            ctx.log_warning(f"Requested resource ({uri}) from unknown source")
            return None

    async def resource_upload(
        self,
        uri: str,
        content: ResourceContent,
        **extra: Any,
    ) -> Meta:
        raise NotImplementedError("Resource uploading is not supported by MCP servers")

    async def resource_delete(
        self,
        uri: str,
        **extra: Any,
    ) -> None:
        raise NotImplementedError("Resource deleting is not supported by MCP servers")

    def _client_identifier_for_uri(
        self,
        uri: str,
        /,
    ) -> str | None:
        """Find server associated with URI."""
        if not uri:
            return None

        parsed: ParseResult = urlparse(uri)
        if not parsed.netloc:
            return None

        if parsed.netloc in self._clients.keys():
            return parsed.netloc

        match parsed.netloc.split(".", 1):
            case [identifier, _]:
                return identifier

            case _:
                return None

    async def tools_fetch(
        self,
        *,
        mcp_server: str | None = None,
        **extra: Any,
    ) -> Sequence[Tool]:
        if mcp_server is None:
            return tuple(
                chain.from_iterable(
                    await gather(*[client.loading(**extra) for client in self._tools.values()])
                )
            )

        elif tools := self._tools.get(mcp_server):
            return await tools.loading(**extra)

        else:
            return ()

    async def __aenter__(self) -> Sequence[ResourcesRepository | ToolsProvider]:
        features: Sequence[Sequence[ResourcesRepository | ToolsProvider]] = await gather(
            *[client.__aenter__() for client in self._clients.values()]
        )

        self._resources = {}
        self._prompts = {}
        self._tools = {}

        for states in features:
            for state in states:
                if isinstance(state, ResourcesRepository):
                    self._resources[cast(str, state.meta["mcp_server"])] = state

                if isinstance(state, ToolsProvider):
                    self._tools[cast(str, state.meta["mcp_server"])] = state

        inherited_features: list[ResourcesRepository | ToolsProvider] = []
        if self._resources:
            inherited_features.append(
                ResourcesRepository(
                    list_fetching=self.resources_list,
                    fetching=self.resource_fetch,
                    uploading=self.resource_upload,
                    deleting=self.resource_delete,
                    meta=Meta({"mcp_server": "mcp_aggregate"}),
                )
            )

        if self._tools:
            inherited_features.append(
                ToolsProvider(
                    loading=self.tools_fetch,
                    meta=Meta({"mcp_server": "mcp_aggregate"}),
                )
            )

        return inherited_features

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await gather(
            *[
                client.__aexit__(
                    exc_type,
                    exc_val,
                    exc_tb,
                )
                for client in self._clients.values()
            ]
        )



from typing import Any

from haiway import getenv_str
from ollama import AsyncClient

__all__ = ("OllamaAPI",)


class OllamaAPI:
    __slots__ = (
        "_client",
        "_server_url",
        "_timeout",
    )

    def __init__(
        self,
        server_url: str | None = None,
        timeout: float | None = None,
        **extra: Any,
    ) -> None:
        # Prefer explicit server_url, fallback to env vars
        self._server_url: str | None = (
            server_url or getenv_str("OLLAMA_SERVER_URL") or getenv_str("OLLAMA_HOST")
        )
        self._timeout: float | None = timeout
        self._client = self._prepare_client()

    def _prepare_client(self) -> AsyncClient:
        return AsyncClient(
            host=self._server_url,
            timeout=self._timeout,
        )

    async def _initialize_client(self) -> None:
        # Recreate and enter async client lifecycle
        try:
            await self._client._client.aclose()
        except Exception:
            # Best-effort close; underlying client may not be initialized yet
            pass  # nosec: B110
        self._client = self._prepare_client()
        await self._client._client.__aenter__()

    async def _deinitialize_client(self) -> None:
        await self._client._client.__aexit__(
            None,
            None,
            None,
        )



from collections.abc import Collection, Iterable
from types import TracebackType
from typing import Any, final

from haiway import State

from draive.embedding import TextEmbedding
from draive.guardrails import GuardrailsModeration
from draive.mistral.api import MistralAPI
from draive.mistral.completions import MistralCompletions
from draive.mistral.embedding import MistralEmbedding
from draive.mistral.moderation import MistralContentModeration
from draive.models import GenerativeModel

__all__ = ("Mistral",)


@final
class Mistral(
    MistralCompletions,
    MistralEmbedding,
    MistralContentModeration,
    MistralAPI,
):
    """
    Access to Mistral services, can be used to prepare various functionalities like lmm.
    """

    __slots__ = ("_features",)

    def __init__(
        self,
        server_url: str | None = None,
        api_key: str | None = None,
        timeout: float | None = None,
        features: Collection[type[GenerativeModel | TextEmbedding | GuardrailsModeration]]
        | None = None,
        **extra: Any,
    ) -> None:
        super().__init__(
            server_url=server_url,
            api_key=api_key,
            timeout=timeout,
            **extra,
        )

        self._features: frozenset[type[State]] = (
            frozenset(features)
            if features is not None
            else frozenset((GenerativeModel, TextEmbedding, GuardrailsModeration))
        )

    async def __aenter__(self) -> Iterable[State]:
        await self._initialize_client()
        state: list[State] = []

        if GenerativeModel in self._features:
            state.append(self.generative_model())

        if TextEmbedding in self._features:
            state.append(self.text_embedding())

        if GuardrailsModeration in self._features:
            state.append(self.content_guardrails())

        return state

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self._deinitialize_client()



from draive.models.generative import GenerativeModel, RealtimeGenerativeModel
from draive.models.instructions import (
    Instructions,
    InstructionsArgumentDeclaration,
    InstructionsDeclaration,
    InstructionsDefining,
    InstructionsListing,
    InstructionsLoading,
    InstructionsMissing,
    InstructionsRemoving,
    InstructionsRepository,
    InstructionsTemplate,
    ResolveableInstructions,
    instructions,
)
from draive.models.tools import (
    FunctionTool,
    ModelToolSpecification,
    Tool,
    ToolAvailabilityChecking,
    Toolbox,
    ToolError,
    ToolErrorFormatting,
    ToolResultFormatting,
    ToolsLoading,
    ToolsProvider,
    ToolsSuggesting,
    tool,
)
from draive.models.types import (
    ModelContext,
    ModelContextElement,
    ModelException,
    ModelGenerating,
    ModelInput,
    ModelInputBlock,
    ModelInputBlocks,
    ModelInputChunk,
    ModelInputInvalid,
    ModelInstructions,
    ModelMemory,
    ModelMemoryRecall,
    ModelOutput,
    ModelOutputBlock,
    ModelOutputBlocks,
    ModelOutputChunk,
    ModelOutputFailed,
    ModelOutputInvalid,
    ModelOutputLimit,
    ModelOutputSelection,
    ModelRateLimit,
    ModelReasoning,
    ModelSession,
    ModelSessionClosing,
    ModelSessionEvent,
    ModelSessionInput,
    ModelSessionOpening,
    ModelSessionOutput,
    ModelSessionOutputSelection,
    ModelSessionPreparing,
    ModelSessionReading,
    ModelSessionScope,
    ModelSessionWriting,
    ModelStreamOutput,
    ModelToolFunctionSpecification,
    ModelToolHandling,
    ModelToolRequest,
    ModelToolResponse,
    ModelToolsDeclaration,
    ModelToolsSelection,
)

__all__ = (
    "FunctionTool",
    "GenerativeModel",
    "Instructions",
    "InstructionsArgumentDeclaration",
    "InstructionsDeclaration",
    "InstructionsDefining",
    "InstructionsListing",
    "InstructionsLoading",
    "InstructionsMissing",
    "InstructionsRemoving",
    "InstructionsRepository",
    "InstructionsTemplate",
    "ModelContext",
    "ModelContextElement",
    "ModelException",
    "ModelGenerating",
    "ModelInput",
    "ModelInputBlock",
    "ModelInputBlocks",
    "ModelInputChunk",
    "ModelInputInvalid",
    "ModelInstructions",
    "ModelMemory",
    "ModelMemoryRecall",
    "ModelOutput",
    "ModelOutputBlock",
    "ModelOutputBlocks",
    "ModelOutputChunk",
    "ModelOutputFailed",
    "ModelOutputInvalid",
    "ModelOutputLimit",
    "ModelOutputSelection",
    "ModelRateLimit",
    "ModelReasoning",
    "ModelSession",
    "ModelSessionClosing",
    "ModelSessionEvent",
    "ModelSessionInput",
    "ModelSessionOpening",
    "ModelSessionOutput",
    "ModelSessionOutputSelection",
    "ModelSessionPreparing",
    "ModelSessionReading",
    "ModelSessionScope",
    "ModelSessionWriting",
    "ModelStreamOutput",
    "ModelToolFunctionSpecification",
    "ModelToolHandling",
    "ModelToolRequest",
    "ModelToolResponse",
    "ModelToolSpecification",
    "ModelToolsDeclaration",
    "ModelToolsSelection",
    "RealtimeGenerativeModel",
    "ResolveableInstructions",
    "Tool",
    "ToolAvailabilityChecking",
    "ToolError",
    "ToolErrorFormatting",
    "ToolResultFormatting",
    "Toolbox",
    "ToolsLoading",
    "ToolsProvider",
    "ToolsSuggesting",
    "instructions",
    "tool",
)



from typing import Any, overload

from haiway import State, statemethod

from draive.guardrails.moderation.types import (
    GuardrailsInputModerationException,
    GuardrailsModerationChecking,
    GuardrailsModerationException,
    GuardrailsOutputModerationException,
)
from draive.multimodal import Multimodal, MultimodalContent

__all__ = ("GuardrailsModeration",)


async def _no_moderation(
    content: MultimodalContent,
    /,
    **extra: Any,
) -> None:
    pass


class GuardrailsModeration(State):
    @overload
    @classmethod
    async def check_input(
        cls,
        content: Multimodal,
        /,
        **extra: Any,
    ) -> None: ...

    @overload
    async def check_input(
        self,
        content: Multimodal,
        /,
        **extra: Any,
    ) -> None: ...

    @statemethod
    async def check_input(
        self,
        content: Multimodal,
        /,
        **extra: Any,
    ) -> None:
        try:
            await self.input_checking(
                MultimodalContent.of(content),
                **extra,
            )

        except GuardrailsModerationException as exc:
            raise GuardrailsInputModerationException(
                f"Input moderation guardrails triggered: {exc}",
                content=exc.content,
                violations=exc.violations,
                replacement=exc.replacement,
            ) from exc

    @overload
    @classmethod
    async def check_output(
        cls,
        content: Multimodal,
        /,
        **extra: Any,
    ) -> None: ...

    @overload
    async def check_output(
        self,
        content: Multimodal,
        /,
        **extra: Any,
    ) -> None: ...

    @statemethod
    async def check_output(
        self,
        content: Multimodal,
        /,
        **extra: Any,
    ) -> None:
        try:
            await self.output_checking(
                MultimodalContent.of(content),
                **extra,
            )

        except GuardrailsModerationException as exc:
            raise GuardrailsOutputModerationException(
                f"Output moderation guardrails triggered: {exc}",
                content=exc.content,
                violations=exc.violations,
                replacement=exc.replacement,
            ) from exc

    input_checking: GuardrailsModerationChecking = _no_moderation
    output_checking: GuardrailsModerationChecking = _no_moderation



from draive.conversation.realtime.state import RealtimeConversation
from draive.conversation.realtime.types import RealtimeConversationSession

__all__ = (
    "RealtimeConversation",
    "RealtimeConversationSession",
)



from typing import cast

from draive.evaluation import EvaluationScore, EvaluationScoreValue, evaluator
from draive.multimodal import Multimodal, MultimodalContent
from draive.stages import Stage

__all__ = ("factual_accuracy_evaluator",)


INSTRUCTION: str = """\
You are evaluating the provided content according to the defined criteria.

<INSTRUCTION>
Carefully examine provided CONTENT, then rate it using solely a\
 factual accuracy metric according to the EVALUATION_CRITERIA.
Think step by step and provide explanation of the score before the final score.
Use the explained RATING scale and the requested FORMAT to provide the result.
</INSTRUCTION>

<EVALUATION_CRITERIA>
Evaluated metric is factual accuracy - the extent to which the content contains\
 factually correct information based on established knowledge, verifiable facts,\
 and reliable sources. This evaluates the correctness of claims, data, statements,\
 and assertions made in the content, regardless of any reference material.\
 The evaluation should consider well-established facts, scientific consensus,\
 historical accuracy, and generally accepted knowledge.
</EVALUATION_CRITERIA>
{guidelines}
<RATING>
Assign a factual accuracy score using exact name of one of the following values:
- "poor" is very low factual accuracy, the content contains many significant\
 factual errors or false information.
- "fair" is low factual accuracy, the content has several factual inaccuracies\
 or questionable claims mixed with some correct information.
- "good" is moderate factual accuracy, the content is mostly factually correct\
 but contains some minor inaccuracies or unverified claims.
- "excellent" is high factual accuracy, the content is largely factually correct\
 with minimal or very minor factual issues.
- "perfect" is very high factual accuracy, the content is completely factually\
 accurate with all information being correct and verifiable.
Use the "none" value for content that cannot be rated at all.
</RATING>

<FORMAT>
The final result containing only the rating value, HAVE to be put inside a `RESULT`\
 xml tag within the result i.e. `<RESULT>good</RESULT>`.
</FORMAT>
"""


@evaluator(name="factual_accuracy")
async def factual_accuracy_evaluator(
    evaluated: Multimodal,
    /,
    guidelines: str | None = None,
) -> EvaluationScore:
    """
    Evaluate factual correctness of content based on established knowledge.

    This evaluator assesses the factual accuracy of content by examining
    claims, data, statements, and assertions against well-established facts,
    scientific consensus, and generally accepted knowledge without requiring
    reference material.

    Parameters
    ----------
    evaluated : Multimodal
        The content to evaluate for factual accuracy
    guidelines : str | None, optional
        Additional guidelines for factual accuracy evaluation, by default None

    Returns
    -------
    EvaluationScore
        Factual accuracy score with categorical rating and explanation

    Raises
    ------
    ValueError
        When the evaluator fails to parse the result
    """
    if not evaluated:
        return EvaluationScore.of(
            0.0,
            meta={"comment": "Input was empty!"},
        )

    completion: MultimodalContent = await Stage.completion(
        MultimodalContent.of(
            "<CONTENT>",
            evaluated,
            "</CONTENT>",
        ),
        instructions=INSTRUCTION.format(
            guidelines=f"\n<GUIDELINES>\n{guidelines}\n</GUIDELINES>\n"
            if guidelines is not None
            else ""
        ),
    ).execute()

    if result := completion.tag("RESULT"):
        return EvaluationScore.of(
            cast(EvaluationScoreValue, result.content.to_str().lower()),
            meta={"comment": completion.to_str()},
        )

    else:
        raise ValueError(f"Invalid evaluator result:\n{completion}")
